/* =====================================================================================================================
 (c) Copyright OMRON Europe BV 2016
 All Right Reserved
 OMRON Development & Engineering
 ================================================================================================================== *//**
 @file

  This file is part of Omron Secure Boot (OSB).

  Omron Secure Boot (OSB) is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Omron Secure Boot (OSB) is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Omron Secure Boot (OSB).  If not, see <http://www.gnu.org/licenses/>.

 Contains app specific logic for unsealing the the secret AES key in TPM. It also implements higher level AES
 encryption/decryption logic.

 @internal
 *//* ================================================================================================================ */

/* ================ Global include files ============================================================================ */
#include "app_aes.h"

#include "tpm_driver.h"
#include "shared.h"
#include "aes.h"
#include "utils.h"
#include "offsets.h"
#include "sha2.h"
#include "app_tpm.h"
#include "convars.h"

/* ================ Sub-system include files ======================================================================== */

/* ================ Local constant definitions ====================================================================== */
#define SEAL_KEY_CONVAR     (VAR_IDENT_STR "_sealedkey")

/* ================ Local macro definitions ========================================================================= */

/* ================ Local type definitions ========================================================================== */

/* ================ Local data definitions ========================================================================== */
static t_bool isInitialized = FALSE;
/**
 * Indicates if a TPM is available in the system
 */
static t_bool tpmAvailable = FALSE;

/**
 * Indicates if the seal key could be retrieved
 */
static t_bool hasSealKey   = FALSE;

/**
 * Array containing the unsealed secret AES key from TPM memory
 */
static t_U8       decKey[AES128_KEYLEN] = { 0 };
static t_U8       sealKey[AES128_KEYLEN] = { 0 };
static const t_U8 iv[AES128_KEYLEN] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const t_U8 mangle[AES128_KEYLEN] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

/* ================ Local functions ================================================================================= */
/**
 *
 * @return
 */
static int lazyLoadKey(void)
{
    int status = OMN_EXIT_SUCCESS;

    if (ISFALSE(hasSealKey)) {
        status = appaes_loadSealKey();
        GRUB_ASSERT_MESSAGE(status == OMN_EXIT_SUCCESS, "Failed to lazy-load the seal AES key");
    }

    if (ISFALSE(isInitialized)) {
        status = appaes_loadAes128DecKey();
        GRUB_ASSERT_MESSAGE(status == OMN_EXIT_SUCCESS, "Failed to lazy-load the secret AES key");
    }

    return status;
}

/**
 *
 * @param input
 * @param output
 * @param length
 */
static void mangleData(t_U8 const * const input, t_U8* const output, t_U32 const length)
{
#if 1
    unsigned idx = 0;

    for (idx = 0; idx < length; idx += sizeof(mangle)) {
        t_U32 const remain = (((length - idx) < sizeof(mangle)) ? length - (t_U32)idx : sizeof(mangle));
        unsigned    bidx;

        for (bidx = 0; bidx < remain; bidx++) {
            output[idx + bidx] = input[idx + bidx] ^ mangle[bidx];
        }
    }
#endif
}

/* ================ Global functions (Sub-system scope) ============================================================= */

/* ================ Global functions (Global system scope) ======================================================= */
/**
 * Load AES key from TPM memory
 * @return exit state indicating if verification key was successfully loaded from TPM memory
 */
int appaes_loadSealKey(void)
{
    int         status = OMN_EXIT_SUCCESS;
    char        pcrIndexStr[AES128_PCRLEN] = { 0 };

    if (ISTRUE(tpmAvailable)) {
        // Read string to TPM
        status = tpm_nvread(TPM_AES_AREA, AES_PCR_OFFSET, (t_U8 * const ) pcrIndexStr, AES128_PCRLEN);

#ifdef DEBUG_AES
        if (status != OMN_EXIT_SUCCESS) {
            printf("Could not read TPM PCR string (status: %d)\n", status);
        }
#endif
        if (status == OMN_EXIT_SUCCESS) {
            t_U8 pcrIndexLen = 32;
            t_U8 pcrIndex[pcrIndexLen];
            t_U8 shaHash[SHA512_DIGEST_LENGTH];
            int  i;

            // write array length
            utils_hexStrToByte(pcrIndexStr, &pcrIndexLen);

            // index string should not be larger then the reserved space in tpm memory
            GRUB_ASSERT_MESSAGE(2*(pcrIndexLen+1) <= AES128_PCRLEN,
                            "TPM PCR length defined for AES key is large then defined AES128_PCRLEN");

            for (i = 0; i < pcrIndexLen; i++) {
                utils_hexStrToByte((pcrIndexStr + i * 2 + 2), &pcrIndex[i]);
            }
            // load key from tpm and decrypt symmetric key with pcr's
            status = apptpm_calculateHashFromPcrs(shaHash, pcrIndex, pcrIndexLen);
            if (status == OMN_EXIT_SUCCESS) {
                GRUB_ASSERT_MESSAGE(SHA512_DIGEST_LENGTH >= sizeof(sealKey), "SHA hash smaller then seal key length")

                (void)memcpy(sealKey, shaHash, sizeof(sealKey));
                hasSealKey = TRUE;

#ifdef DEBUG
                printf("seal key: ");
                utils_printKeyAsHexStr(sealKey, sizeof(sealKey));
                printf("\n\n");
#endif
            }
        }
    } else {
#ifdef DEBUG
        printf("No TPM available to calculate seal-key\n");
#endif
        (void)memset(sealKey, 0, sizeof(sealKey));
        status = OMN_EXIT_FAILURE;
    }

    return status;
}

/**
 * Load AES key from TPM memory
 * @return exit state indicating if verification key was successfully loaded from TPM memory
 */
int appaes_loadAes128DecKey(void)
{
    int         status = OMN_EXIT_SUCCESS;

    if (ISFALSE(hasSealKey)) {
        status = appaes_loadSealKey();
    }
    if (ISTRUE(tpmAvailable) && (status == OMN_EXIT_SUCCESS)) {
        char*           convarValue;
        T_CONVAR_RESULT convarResult = convars_getVar(SEAL_KEY_CONVAR, &convarValue);

        if (convarResult == CONVAR_OK) {
            t_U8      decKeyEnc[AES128_KEYLEN] = { 0 };
            int const sealedKeyLength          = strlen(convarValue);

            if (sealedKeyLength == (sizeof(decKeyEnc)*2)) {
                t_U32 decKeyLength = sizeof(decKeyEnc);

                (void)utils_hexStrToArray(decKeyEnc, &decKeyLength, (char const * const)convarValue);

                GRUB_ASSERT_MESSAGE((decKeyLength == AES128_KEYLEN), "Sealkey length should be AES128_KEYLEN");

#ifdef DEBUG_AES
                printf("sealed AES key: ");
                utils_printKeyAsHexStr(decKeyEnc, sizeof(decKeyEnc));
                printf("\n");
#endif

                aes_ecb_decrypt(decKeyEnc, sealKey, decKey);

#ifdef DEBUG_AES
                printf("unsealed AES key: ");
                utils_printKeyAsHexStr(decKey, sizeof(decKey));
                printf("\n\n");
#endif
            } else {
                (void)memset(decKey, 0, sizeof(decKey));
                (void)convars_defVarBin(SEAL_KEY_CONVAR, decKey, sizeof(decKey), FALSE, TRUE);
                status = OMN_EXIT_FAILURE;
            }
        } else {
            (void)memset(decKey, 0, sizeof(decKey));
            (void)convars_defVarBin(SEAL_KEY_CONVAR, decKey, sizeof(decKey), FALSE, TRUE);
            status = OMN_EXIT_FAILURE;
        }
    } else {
        status = OMN_EXIT_FAILURE;
    }

    return status;
}

/**
 * @param input     [in]  buffer of data to be decrypted
 * @param length    [in]  The size of the input buffer; should be @p AES128_KEYLEN size
 * @param output    [out] verified data (decrypted with symmetric key)
 * @return exit state indicating if verification was successfully
 */
int appaes_aes128DecryptEcb(void const * const input, void * const output, t_U32 const length)
{
    int status = lazyLoadKey();

    if (status == OMN_EXIT_SUCCESS) {
        if (length == AES128_KEYLEN) {
            aes_ecb_decrypt(input, decKey, output);
        } else {
            status = OMN_EXIT_FAILURE;
        }
    }
    return status;
}

/**
 * Seals the @p input with the seal-key and write the result to @p output
 * @param input     [in]  buffer of data to be decrypted
 * @param length    [in]  The size of the input buffer; should be @p AES128_KEYLEN aligned
 * @param output    [out] sealed data
 * @return exit state indicating if verification was successfully
 */
int appaes_seal(void* const input, void * const output, t_U32 const length)
{
    int status = lazyLoadKey();

    if (status == OMN_EXIT_SUCCESS) {
        if (((length % AES128_KEYLEN) == 0) && (status == OMN_EXIT_SUCCESS)) {
            aes_cbc_encrypt(output, input, length, sealKey, (t_U8* const)iv);
            status = OMN_EXIT_SUCCESS;
        } else {
            status = OMN_EXIT_FAILURE;
        }
    }
    return status;
}

/**
 * Provides the seal-key to the caller
 * Be VERY careful with this!! It should be a secret
 * @param keySize   [out] The size, in bytes, of the seal-key
 * @return A pointer to the seal-key. Make sure you do NOT change the contents!
 */
t_U8* appaes_getSealKey(t_U16* const keySize)
{
    t_U8* result = NULL;
    int   status = lazyLoadKey();

    if (status == OMN_EXIT_SUCCESS) {
        result = sealKey;
        if (keySize != NULL) {
            *keySize = sizeof(sealKey);
        }
    }
    return result;
}

/**
 * @param input     [in]  buffer of data to be decrypted
 * @param length    [in]  The size of the input buffer
 * @param output    [out] verified data (decrypted with symmetric key)
 * @return exit state indicating if verification was successfully
 */
int appaes_aes128DecryptCbc(void* const input, void * const output, t_U32 const length)
{
    int const status = lazyLoadKey();

    if (status == OMN_EXIT_SUCCESS) {
        aes_cbc_decrypt(output, input, length, decKey, (t_U8* const)iv);
    }
    return status;
}

/**
 * Clears the key from memory
 */
void appaes_destroyKey(void)
{
    (void)memset(decKey, 0, sizeof(decKey)); //TODO Better is to do it randomly so an attacker doesn't see all zero's
    (void)memset(sealKey, 0, sizeof(sealKey));
    (void)convars_defVarBin(SEAL_KEY_CONVAR, decKey, sizeof(decKey), FALSE, TRUE);
    hasSealKey    = FALSE;
    isInitialized = FALSE;
}

/**
 * Initializes the module
 * @param hasTpm [in] A boolean indicating if a TPM is available (@p TRUE) or not (@p FALSE)
 */
void appaes_init(t_bool const hasTpm)
{
    int status;

    (void)memset(decKey, 0, sizeof(decKey));
    (void)convars_defVarBin(SEAL_KEY_CONVAR, decKey, sizeof(decKey), FALSE, TRUE);
    tpmAvailable = hasTpm;
    if (ISTRUE(tpmAvailable)) {
        status = tpm_nvread(TPM_AES_AREA, AES_KEY_OFFSET, decKey, AES128_KEYLEN);
        if (status != OMN_EXIT_SUCCESS) {
            (void)memset(decKey, 0, sizeof(decKey));
        } else {
            mangleData(decKey, decKey, sizeof(decKey)); //Unmangle
            isInitialized = TRUE;
        }
        status = appaes_loadSealKey();
        if (status != OMN_EXIT_SUCCESS) {
            (void)memset(sealKey, 0, sizeof(sealKey));
        }
    }
}
/* ================================================================================================================ */

/* ================ End of file ===================================================================================== */
