/* =====================================================================================================================
 (c) Copyright OMRON Europe BV 2016
 All Right Reserved
 OMRON Development & Engineering
 ================================================================================================================== *//**
 @file

  This file is part of Omron Secure Boot (OSB).

  Omron Secure Boot (OSB) is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Omron Secure Boot (OSB) is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Omron Secure Boot (OSB).  If not, see <http://www.gnu.org/licenses/>.

 Implements the application logic for storing the RSA key in the TPM and to retrieve it when verifying data

 @internal
 *//* ================================================================================================================ */

/* ================ Global include files ============================================================================ */
#include "app_rsa.h"

#include "tpm_driver.h"
#include "rsa.h"
#include "utils.h"
#include "bigdigits.h"
#include "rsa_cfg.h"
#include "offsets.h"

/* ================ Sub-system include files ======================================================================== */

/* ================ Local constant definitions ====================================================================== */

/* ================ Local macro definitions ========================================================================= */

/* ================ Local type definitions ========================================================================== */

/* ================ Local data definitions ========================================================================== */

// public encrypt or verification key
static DIGIT_T e[RSA_KEYLEN / BYTES_PER_DIGIT] = { 0 };

// modulus n=p*q for prime numbers p and q
static DIGIT_T n[RSA_KEYLEN / BYTES_PER_DIGIT] = { 0 };

// Indication if RSA keys are already loaded from tpm
static t_bool isInitialized = FALSE;

static int keyIndex = 0;
/* ================ Local functions ================================================================================= */
/**
 * Converts the linear RSA key index (0...n) to an absolute index in a section
 * Because the TPM can only have 2000 Bytes in a linear address space, multiple sections are created. Each contain three RSA key-pairs.
 * This function will convert the linear index (0...n) to a section and relative offset inside that section.
 * For example: - index = 2 -> first section with index 2
 *              - index = 3 -> Second section with index 0
 *
 * @param index [inout] In: The linear key-index; Out: The relative index inside the returned section
 * @return The TPM section identifier belonging to the linear key index
 */
static unsigned getTpmAreaFromIndex(int* const index)
{
    int result;

    if ((*index >= 0) && (*index < 3)) {
        result = TPM_RSA0_2_AREA;
    } else if ((*index >= 3) && (*index < 6)) {
        result = TPM_RSA3_5_AREA;
        *index = *index - 3;
    } else {
        result = 0;
        *index = -1;
    }

    return result;
}

/**
 * Load RSA keys from TPM memory
 * @param rsaKeyIndex The index to load the RSA keys from
 * @return boolean indicating if RSA key is successfully loaded in TPM
 */
static int loadRsa(int rsaKeyIndex)
{
    int            status  = OMN_EXIT_FAILURE;
    unsigned const tpmArea = getTpmAreaFromIndex(&rsaKeyIndex);

    if ((rsaKeyIndex >= 0) && (rsaKeyIndex < NR_RSA_KEYS)) {
        t_U32 const eOffset = (t_U32)(RSA_KEY_E_OFFSET+(RSA_KEYLEN*rsaKeyIndex*2));
        t_U32 const nOffset = (t_U32)(RSA_KEY_E_OFFSET+(RSA_KEYLEN*rsaKeyIndex*2)+RSA_KEYLEN);

        // Read public sign key to tpm memory
        status = tpm_nvread(tpmArea, eOffset, (t_U8* const ) e, RSA_KEYLEN);
        if (status == OMN_EXIT_SUCCESS) {
            // Read modulus to tpm memory
            status = tpm_nvread(tpmArea, nOffset, (t_U8* const ) n, RSA_KEYLEN);

            if (status == OMN_EXIT_SUCCESS) {
                isInitialized = TRUE;
    #ifdef DEBUG_RSA
                printf("RSA keys loaded from index %d:\n e: ", rsaKeyIndex);
                utils_printKeyAsHexStr((t_U8 *) e, RSA_KEYLEN);
                printf("\n n: ");
                utils_printKeyAsHexStr((t_U8 *) n, RSA_KEYLEN);
                printf("\n\n");
    #endif
            }
        }
    }
    return status;
}

/* ================ Global functions (Sub-system scope) ============================================================= */

/* ================ Global functions (Global system scope) ======================================================= */

/**
 * Verify signed data
 * @param input buffer of data to be verified
 * @param inputLen [in] length of input buffer in bytes
 * @param output verified data (encrypted with public key)
 * @param outputLen [inout] In: Maximum length of output buffer in bytes Out: length of output buffer in bytes
 * @return EXIT code indicating if verification was successfully
 */
int apprsa_rsaVerify(t_U8* const input, int const inputLen, t_U8* const output, int* const outputLen)
{
    int status = OMN_EXIT_SUCCESS;
    if (ISFALSE(isInitialized)) {
        status = loadRsa(keyIndex);
    }

    if (status == OMN_EXIT_SUCCESS) {
        status = rsa_verify(n, e, input, inputLen, output, outputLen);
    }
    return status;
}

/**
 * Clears the key from memory
 */
void apprsa_destroyKey(void)
{
    (void) memset(n, 0, sizeof(n)); //TODO Better is to do it randomly so an attacker doesn't see all zero's
    (void) memset(e, 0, sizeof(e)); //TODO Better is to do it randomly so an attacker doesn't see all zero's
    isInitialized = FALSE;
}

/**
 *
 * @return
 */
int apprsa_resetKeyIndex(void)
{
    keyIndex      = 0;
    isInitialized = FALSE;
    return keyIndex;
}

/**
 * Update key index
 * @return void
 */
void apprsa_setKeyIndex(int const newKeyIndex)
{
    keyIndex = newKeyIndex;
}

/**
 *
 * @return
 */
int apprsa_nextKeyIndex(void)
{
    if (keyIndex >= 0) {
        keyIndex++;
        isInitialized = FALSE;
    }

    if (keyIndex >= NR_RSA_KEYS) {
        keyIndex = -1;
    }
    return keyIndex;
}
/* ================ End of file ===================================================================================== */
