/* =====================================================================================================================
 (c) Copyright OMRON Europe BV 2016
 All Right Reserved
 OMRON Development & Engineering
 ================================================================================================================== *//**
 @file

  This file is part of Omron Secure Boot (OSB).

  Omron Secure Boot (OSB) is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Omron Secure Boot (OSB) is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Omron Secure Boot (OSB).  If not, see <http://www.gnu.org/licenses/>.

 Implements the application logic for interaction with the TPM

 @internal
 *//* ================================================================================================================ */

/* ================ Global include files ============================================================================ */
#include "app_tpm.h"

#include "tpm_driver.h"
#include "shared.h"
#include "aes.h"
#include "sha2.h"
#include "utils.h"
#include "offsets.h"

/* ================ Sub-system include files ======================================================================== */

/* ================ Local constant definitions ====================================================================== */

/* ================ Local macro definitions ========================================================================= */

/* ================ Local type definitions ========================================================================== */

/* ================ Local data definitions ========================================================================== */

/* ================ Local functions ================================================================================= */

/* ================ Global functions (Sub-system scope) ============================================================= */

/* ================ Global functions (Global system scope) ======================================================= */

/**
 * Calculate HASH of TPM PCR's
 * @param shaHash      [out] hash to be calculated
 * @param pcrindex      [in]  array of PCR index values
 * @param pcrindexLen   [in]  length of PCR array
 * @return exit state indicating if calculation of hash was successful
 */
int apptpm_calculateHashFromPcrs(t_U8 * const shaHash, t_U8 const * const pcrindex, int const pcrindexLen)
{
    int status = OMN_EXIT_SUCCESS;

    if (shaHash == NULL) {
        status = OMN_EXIT_FAILURE;
    } else {
        SHA512_CTX shaCtx;
        SHA512_Init(&shaCtx);
        int i;
        for (i = 0; i < pcrindexLen; i++) {
            t_U8 pcrValue[PCR_SIZE];
            // read pcr
            status = tpm_readPcrObscure(pcrindex[i], pcrValue);
            if (status != OMN_EXIT_SUCCESS) {
                break;
            }
#ifdef DEBUG_TPM
            printf("PCR%d: ", pcrindex[i]);
            utils_printAsHexStr(pcrValue, PCR_SIZE);
            printf("\n");
#endif
            SHA512_Update(&shaCtx, pcrValue, PCR_SIZE);
        }

        if (status == OMN_EXIT_SUCCESS) {
            t_U32 shaHash32[SHA2_DWORD_SIZE];
            SHA512_Final((sha2_byte*) shaHash32, &shaCtx);
            memcpy(shaHash, shaHash32, SHA512_DIGEST_LENGTH);
#ifdef DEBUG_TPM
            printf("SHA: ");
            utils_printAsHexStr(shaHash, SHA512_DIGEST_LENGTH);
            printf("\n");
#endif
        }
    }
    return status;
}

/**
 * Mangle all TPM measurements
 * @return exit state indicating if measurement mix was successful
 */
int apptpm_mixPcrs(void)
{
    int status = OMN_EXIT_SUCCESS;
    t_U8 pcrIndex;
    t_U32 const dataSize = PCR_SIZE;
    t_U8 data[dataSize];
    for (pcrIndex = 0; pcrIndex < TPM_PCRS; pcrIndex++) {
        t_U8 i;
        // fill data with random values
        for (i = 0; i < dataSize; i++) {
            data[i] = (t_U8) (utils_pseudoRand() % 256U);
        }

        // extend tpm with measurement
        status = tpm_extendPcr(pcrIndex, data, dataSize);
        if (status == OMN_EXIT_FAILURE) {
            break;
        }
    }
    return status;
}

/* ================================================================================================================ */

/* ================ End of file ===================================================================================== */
