/* =====================================================================================================================
 (c) Copyright OMRON Europe BV 2016
 All Right Reserved
 OMRON Development & Engineering
 ================================================================================================================== *//**
 @file
 This file is part of Omron Secure Boot (OSB).

 Omron Secure Boot (OSB) is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Omron Secure Boot (OSB) is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Omron Secure Boot (OSB). If not, see <http://www.gnu.org/licenses/>.

 CMOS interface to acquire system time

 @internal
 *//* ================================================================================================================ */

/* ================ Global include files ============================================================================ */

#include "bios.h"
#include "shared.h"

/* ================ Sub-system include files ======================================================================== */

/* ================ Local constant definitions ====================================================================== */

#define CMOS_ADDR          ((t_U8)0x70)
#define CMOS_DATA          ((t_U8)0x71)
#define CMOS_REG_SECONDS   ((t_U8)0x00)
#define CMOS_REG_MINUTES   ((t_U8)0x02)
#define CMOS_REG_HOURS     ((t_U8)0x04)

/**
 * timer related status register A
 */
#define CMOS_STATUS_REG_A  ((t_U8)0x0A)

/**
 * date and alarm related status register B
 */
#define CMOS_STATUS_REG_B  ((t_U8)0x0B)

/**
 * date-mode bit specifies whether the internal counters use binary-coded-decimal (BCD) or binary format for
 * time-and-date calendar updates. When this bit is set to 1, the binary format is used. The system initializes
 * this bit to 0.
 */
#define CMOS_DM_BIT        ((t_U8)0x04)

/**
 * 24/12-hour bit specifies whether the hour byte is in 12-hour or 24-hour mode. When this bit is set to 1, the
 * 24-hour mode is used. The system initializes this bit to 1.
 */
#define CMOS_24H_12H_BIT   ((t_U8)0x02)

#define SECONDS_IN_DAY     ((t_U8)24*60*60)
#define HOURS_IN_DAY       ((t_U8)24)

/* ================ Local macro definitions ========================================================================= */

/* ================ Local type definitions ========================================================================== */

/* ================ Local data definitions ========================================================================== */

/**
 * Variable to store time end time in seconds
 */
static t_U32 timeEnd;

/**
 * Variable to store timer start time in seconds
 */
static t_U32 timeStart;

/* ================ Local functions ================================================================================= */

static t_U8 cmos_in8(t_U16 const port)
{
    t_U8 data;

    asm volatile ( "inb %[port], %[data]" : [data] "=a"(data) : [port] "Nd"(port) );

    return data;
}

static void cmos_out8(t_U16 const port, t_U8 const data)
{
    asm volatile ( "outb %[data], %[port]" : : [data] "a"(data), [port] "Nd"(port) );
}

static t_bool cmos_isUpdateInProgress(void)
{
	t_U8 statusFlags;

    cmos_out8(CMOS_ADDR, CMOS_STATUS_REG_A);

    //  Bit 7 indicates0: Date and time can be read, 1: Time update in progress
    statusFlags = cmos_in8(CMOS_DATA);
    if ((statusFlags & (1U << 7U)) == 0) {
        return FALSE;
    }
    return TRUE;
}

static t_U8 cmos_getRegistry(t_U8 const reg)
{
	t_U8 data;

    cmos_out8(CMOS_ADDR, reg);
    data = cmos_in8(CMOS_DATA);
    return data;
}

static void cmos_readHMS(t_U8* const second, t_U8* const minute, t_U8* const hour)
{
	t_bool updateInProgress;

    // simulate atomic read and make sure an update isn't in progress before and after reading the registers
    do {
        do {
            updateInProgress = cmos_isUpdateInProgress();
        } while (ISTRUE(updateInProgress));

        *second = cmos_getRegistry(CMOS_REG_SECONDS);
        *minute = cmos_getRegistry(CMOS_REG_MINUTES);
        *hour = cmos_getRegistry(CMOS_REG_HOURS);

        updateInProgress = cmos_isUpdateInProgress();
    } while (ISTRUE(updateInProgress));
}

static void cmos_readSecondsOfDay(t_U32* time)
{
    t_U8 second;
    t_U8 minute;
    t_U8 hour;
    t_U8 lastSecond;
    t_U8 lastMinute;
    t_U8 lastHour;
    t_U8 regB;

    // consecutive reads should result in the same values
    cmos_readHMS(&second, &minute, &hour);
    do {
        lastSecond = second;
        lastMinute = minute;
        lastHour = hour;
        cmos_readHMS(&second, &minute, &hour);
    } while ((lastSecond != second) || (lastMinute != minute) || (lastHour != hour));

    regB = cmos_getRegistry(CMOS_STATUS_REG_B);

    // convert BCD to binary values if necessary
    if ((regB & CMOS_DM_BIT) == 0) {
        second = (second & 0x0F) + ((second / 16U) * 10U);
        minute = (minute & 0x0F) + ((minute / 16U) * 10U);
        hour = ((hour & 0x0F) + (((hour & 0x70U) / 16U) * 10U)) | (hour & 0x80U);
    }

    // convert 12 hour clock to 24 hour clock if necessary
    if (((regB & CMOS_24H_12H_BIT) == 0) && ((hour & 0x80U) != 0)) {
        hour = ((hour & 0x7F) + 12U) % HOURS_IN_DAY;
    }

    // convert to seconds in day
    *time = second + (minute + hour * 60U) * 60U;
}

static void cmos_AddSecondsOfDay(t_U32* const time, t_U32 const timeAdd)
{
    *time = (*time + timeAdd) % SECONDS_IN_DAY;
}

/* ================ Global functions (Sub-system scope) ============================================================= */

/* ================ Global functions (Global system scope) ======================================================= */

/**
 * Initialize timer
 * @param delay [in] delay in seconds ranging from 1 to SECONDS_IN_DAY-1
 * @return void
 */
void bios_timerInit(t_U32 const delay)
{
    cmos_readSecondsOfDay(&timeStart);
    timeEnd = timeStart;
    cmos_AddSecondsOfDay(&timeEnd, delay);
}

/**
 * Check if timer has overflowed
 * @param delay [in] delay in seconds ranging from 1 to SECONDS_IN_DAY-1
 * @return exit status indicating if delay operation was successful
 */
int bios_timerOverflow(void)
{
    t_U32 timeNow;
    cmos_readSecondsOfDay(&timeNow);
    if (timeStart <= timeEnd) {
        // ___ Start --- Now --- End ___
        if ((timeNow >= timeEnd) || (timeNow < timeStart)) {
            return TRUE;
        } else {
            return FALSE;
        }
    } else {
        // Now --- End ___ Start --- Now
        if ((timeNow <= timeEnd) || (timeNow > timeStart)) {
            return TRUE;
        } else {
            return FALSE;
        }
    }
}

/* ================================================================================================================ */

/* ================ End of file ===================================================================================== */
