/* =====================================================================================================================
 (c) Copyright OMRON Europe BV 2016
 All Right Reserved
 OMRON Development & Engineering
 ================================================================================================================== *//**
 @file

 This file is part of Omron Secure Boot (OSB).

 Omron Secure Boot (OSB) is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Omron Secure Boot (OSB) is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Omron Secure Boot (OSB).  If not, see <http://www.gnu.org/licenses/>.

 Partial implementation of SM BIOS, implementation according to system management BIOS reference specification <http://www.dmtf.org/sites/default/files/standards/documents/DSP0134_2.7.1.pdf>

 @internal
 *//* ================================================================================================================ */

/* ================ Global include files ============================================================================ */

#include "bios.h"
#include "types.h"
#include "shared.h"
#include "utils.h"

/* ================ Sub-system include files ======================================================================== */

/* ================ Local constant definitions ====================================================================== */
#define DMI_MEM_START   (0xF0000)
#define DMI_MEM_END     (0x100000)
#define DMI_ENTRY_STR   ("_DMI_")
#define DMI_ENTRY_STR_LEN (5)
#define DMI_ADDR_WIDTH (0x02)

#define SM_ENTRY_STR    ("_SM_")
#define SM_ENTRY_STR_LEN (4)
#define SM_FORMATTED_AREA_LEN (5)
#define SM_ADDR_WIDTH (0x10)

/* ================ Local macro definitions ========================================================================= */

/* ================ Local type definitions ========================================================================== */

typedef struct s_dmiBlock {
    t_U8 type;
    t_U8 length;
    t_U16 handle;
    t_U8 data[];
}PACKED_BYTE_ALIGN(T_DMI_BLOCK);

typedef struct s_dmiEntry {
    char entryStr[DMI_ENTRY_STR_LEN];
    t_U8 checksum;
    t_U16 tableLen;
    t_U32 tableAddr;
    t_U16 blockCnt;
    t_U8 bcdRevision;
}PACKED_BYTE_ALIGN(T_DMI_ENTRY);

typedef struct s_smbiosEntry {
    char entryStr[SM_ENTRY_STR_LEN];
    t_U8 checksum;
    t_U8 tableLength;
    t_U8 majorVersion;
    t_U8 minorVersion;
    t_U16 maxStructureSize;
    t_U8 entryPointRevision;
    char FormattedArea[SM_FORMATTED_AREA_LEN];
    struct s_dmiEntry dmiEntry;
}PACKED_BYTE_ALIGN(T_SMBIOS_ENTRY);

/* ================ Local data definitions ========================================================================== */

/* ================ Local functions ================================================================================= */

/**
 * Verify buffer by calculating the checksum
 * @param buf    buffer to validate
 * @param bufLen length of the buffer in bytes
 * @return returns true if validation was successful
 */
static t_bool validateChecksum(t_U8 const* const buf, int const bufLen)
{
    t_U8 sum = 0;
    int i;
    for (i = 0; i < bufLen; i++) {
        sum += buf[i];
    }
    if (sum == 0) {
        return TRUE;
    } else {
        return FALSE;
    }
}

/**
 * search DMI entries containing specific data
 * @param memAddr     DMI table start
 * @param memAddrLen  DMI table length in bytes
 * @param entryMax    Number of entries
 * @param dmiType     DMI type we are looking for
 * @param dataOffset  DMI offset of data that we want to compare
 * @param dataLen     Length of data
 * @param data        data we want to compare
 * @return returns TRUE if DMI entry was found
 */
static t_bool searchDmiEntries(char const* memAddr,
                               t_U16 const memAddrLen,
                               t_U16 const entryMax,
                               t_U8 const dmiType,
                               t_U32 const dataOffset,
                               t_U32 const dataLen,
                               t_U8 const* const data)
{
    t_U16 entry = 0;
    t_bool entryFound = FALSE;

    // search all DMI entry indices while no valid entry has been found
    while ((entry < entryMax) && (ISFALSE(entryFound))) {
        T_DMI_BLOCK const* dmiBlock = (T_DMI_BLOCK*) memAddr;

        // check if we are inside memory rnage
        GRUB_ASSERT_MESSAGE(memAddr + dmiBlock->length <= memAddr + memAddrLen, "DMI memory index out of range");

        // check if the entry type is correct
        if (dmiBlock->type == dmiType) {
#ifdef DEBUG
            printf("DMI-%d: ", dmiBlock->type);
            utils_printAsHexStr((t_U8 const* const ) dmiBlock, dmiBlock->length);
            printf("\n");
#endif

            // check if the input data and offset is in range
            if (dataLen + dataOffset <= dmiBlock->length) {
#ifdef DEBUG
                printf("(dmi data)");
                utils_printAsHexStr((t_U8 const* const ) dmiBlock + dataOffset, dataLen);
                printf("==(input data)");
                utils_printAsHexStr((t_U8 const* const ) data, dataLen);
                printf("\n");
#endif

                if (memcmp((const char *) (memAddr + dataOffset), (const char *) data, dataLen) == 0) {
                    entryFound = TRUE;
                    break;
                }
            }
        }

        // search next entry
        memAddr = memAddr + dmiBlock->length;
        while (*(t_U8*) (memAddr) != 0 || *(t_U8*) (memAddr + 1) != 0) {
            // skip zero padding
            memAddr++;
        }

        // increase index and memory pointer
        memAddr += DMI_ADDR_WIDTH;
        entry++;
    }

    return entryFound;
}

/* ================ Global functions (Sub-system scope) ============================================================= */

/* ================ Global functions (Global system scope) ======================================================= */

/**
 * Detect SM BIOS and search for DMI entries containing specific data
 * @param dmiType     DMI type we are looking for
 * @param dataOffset  DMI offset of data that we want to compare
 * @param dataLen     Length of data in bytes
 * @param data        data we want to compare
 * @return returns TRUE if DMI entry was found
 */
t_bool bios_cmpDmiEntry(t_U8 const dmiType, t_U32 const dataOffset, t_U32 const dataLen, t_U8 const* const data)
{
    char *memAddr = (char *) DMI_MEM_START;
    t_bool entryFound = FALSE;

    // search entire address space while no valid entry has been found
    while (((t_U32) memAddr < DMI_MEM_END) && (ISFALSE(entryFound))) {
        // search for valid SM BIOS entry
        T_SMBIOS_ENTRY* smEntry = (T_SMBIOS_ENTRY*) memAddr;
        int cmpResult = memcmp(smEntry->entryStr, SM_ENTRY_STR, sizeof(smEntry->entryStr));
        if (cmpResult == 0) {
            t_bool isValid = validateChecksum((t_U8 const* const ) smEntry, sizeof(T_SMBIOS_ENTRY));

            if (ISTRUE(isValid)) {
#ifdef DEBUG
                printf("SM BIOS detected\n", smEntry);
#endif
                // validate valid DMI entry header
                T_DMI_ENTRY const* const dmiEntry = &(smEntry->dmiEntry);
                cmpResult = memcmp(dmiEntry->entryStr, DMI_ENTRY_STR, sizeof(dmiEntry->entryStr));
                if (cmpResult == 0) {
                    isValid = validateChecksum((t_U8 const* const ) dmiEntry, sizeof(T_DMI_ENTRY));
                    if (ISTRUE(isValid)) {
                        // search all dmi entries for data
                        entryFound = searchDmiEntries((void const* const ) (dmiEntry->tableAddr), dmiEntry->tableLen, dmiEntry->blockCnt,
                                        dmiType, dataOffset, dataLen, data);
                    }
                }
            }
        }
        // increase memory pointer
        memAddr += SM_ADDR_WIDTH;
    }

    return entryFound;
}

/* ================ End of file ===================================================================================== */

