/* =====================================================================================================================
 (c) Copyright OMRON Europe BV 2016
 All Right Reserved
 OMRON Development & Engineering
 ================================================================================================================== *//**
 @file

  This file is part of Omron Secure Boot (OSB).

  Omron Secure Boot (OSB) is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Omron Secure Boot (OSB) is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Omron Secure Boot (OSB).  If not, see <http://www.gnu.org/licenses/>.

 Contains logic to implement RSA encryption (create) and decryption (verify) using the bigdigits library

 @internal
 *//* ================================================================================================================ */

/* ================ Global include files ============================================================================ */
#include "rsa.h"
#include "utils.h"

/* ================ Sub-system include files ======================================================================== */

/* ================ Local constant definitions ====================================================================== */

/* ================ Local macro definitions ========================================================================= */

/* ================ Local type definitions ========================================================================== */

/* ================ Local data definitions ========================================================================== */

/* ================ Local functions ================================================================================= */
// We put this here because sometimes we use glibc, sometimes we don't
extern int rand(void);

/**
 * Fills a buffer with random data
 * @param buf       [out] Buffer to receive random data
 * @param bufSize   [in]  The size of the buffer
 */
static void randomFill(t_U8* buf, int bufSize) {
    while (bufSize > 0) {
        *buf = (t_U8) rand();
        buf++;
        bufSize--;
    }
}

/**
 * Generic function to encrypt/decrypt (create/verify) arbitrary data that should be smaller then the used modulus
 * @param mod       [in]  The modulus of the RSA function
 * @param key       [in]  The private/public RSA key
 * @param input     [in]  Arbitrary input data
 * @param inputLen  [in]  The length in bytes of the input data
 * @param output    [out] A pointer to an array for receiving the encrypted/decrypted data (should be at least
 *                        ((@p inputLen + @p AES_KEY_lEN-1) / @p AES_KEY_KEN) * @p AES_KEY_LEN in size)
 * @param outputLen [inout] In: The size of the output-buffer in bytes; Out: The actual length of data in @p output
 *                          after encryption/decryption
 * @return 0 when successful.
 */
static int rsa_exec(DIGIT_T* const      mod,
                    DIGIT_T* const      key,
                    t_U8 const * const  input,
                    int const           inputLen,
                    t_U8* const         output,
                    int* const          outputLen) {
    int cmp;
    int status = OMN_EXIT_SUCCESS;

    t_U8 m[CEIL((CEIL(inputLen, RSA_KEYLEN)*RSA_KEYLEN), BYTES_PER_DIGIT) * BYTES_PER_DIGIT];
    t_U8 s[CEIL((CEIL(inputLen, RSA_KEYLEN)*RSA_KEYLEN), BYTES_PER_DIGIT) * BYTES_PER_DIGIT];

    // generate random noise to fill the empty space
    randomFill(s, sizeof(s)-1);
    s[sizeof(s)-1] = 0;

    // copy input to key
    (void) memcpy(s, input, inputLen);

    // check that our message is not larger or equal to the modulus we are using
    cmp = mpCompare((DIGIT_T *)s, mod, (sizeof(s) / BYTES_PER_DIGIT));
    if (cmp != -1) status = OMN_EXIT_FAILURE;

    if (status == OMN_EXIT_SUCCESS) {
        // calculate exponent modulus of message m to create signed message s given public sign key e modulus n
        int ret = mpModExp((DIGIT_T*) m, (DIGIT_T*) s, key, mod, sizeof(s) / BYTES_PER_DIGIT);
        if (ret == 0) {
            if (*outputLen >= sizeof(m)) {
                *outputLen = sizeof(m);
                (void) memcpy(output, m, sizeof(m));
            } else {
                status = OMN_EXIT_FAILURE;
            }
        } else {
            status = OMN_EXIT_FAILURE;
        }
    }

    return status;
}

/* ================ Global functions (Sub-system scope) ============================================================= */

/* ================ Global functions (Global system scope) ======================================================= *//**
 @endinternal
 *//* ================================================================================================================ */
/**
 * Function to create/decrypt arbitrary data that should be smaller then the used modulus
 * @param mod       [in]  The modulus of the RSA function
 * @param key       [in]  The private RSA key
 * @param input     [in]  Arbitrary input data
 * @param inputLen  [in]  The length in bytes of the input data
 * @param output    [out] A pointer to an array for receiving the encrypted/decrypted data (should be at least
 *                        ((@p inputLen + @p AES_KEY_lEN-1) / @p AES_KEY_KEN) * @p AES_KEY_LEN in size)
 * @param outputLen [inout] In: The size of the output-buffer in bytes; Out: The actual length of data in @p output
 *                          after encryption
 * @return 0 when successful.
 */
int rsa_create(DIGIT_T* const      mod,
               DIGIT_T* const      key,
               t_U8 const * const  input,
               int const           inputLen,
               t_U8* const         output,
               int* const          outputLen)
{
    return rsa_exec(mod, key, input, inputLen, output, outputLen);
}

/**
 * Function to verify/encrypt arbitrary data that should be smaller then the used modulus
 * @param mod       [in]  The modulus of the RSA function
 * @param key       [in]  The public RSA key
 * @param input     [in]  Arbitrary input data
 * @param inputLen  [in]  The length in bytes of the input data
 * @param output    [out] A pointer to an array for receiving the encrypted/decrypted data (should be at least
 *                        ((@p inputLen + @p AES_KEY_lEN-1) / @p AES_KEY_KEN) * @p AES_KEY_LEN in size)
 * @param outputLen [inout] In: The size of the output-buffer in bytes; Out: The actual length of data in @p output
 *                          after encryption
 * @return 0 when successful.
 */
int rsa_verify(DIGIT_T* const      mod,
               DIGIT_T* const      key,
               t_U8 const * const  input,
               int const           inputLen,
               t_U8* const         output,
               int* const          outputLen)
{
    return rsa_exec(mod, key, input, inputLen, output, outputLen);
}

/* ================ End of file ===================================================================================== */
