/* =====================================================================================================================
 (c) Copyright OMRON Europe BV ${year}
 All Right Reserved
 OMRON Development & Engineering
 =====================================================================================================================*/
/**
  @file tpm.c

  This file is part of Omron Secure Boot (OSB).

  Omron Secure Boot (OSB) is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Omron Secure Boot (OSB) is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Omron Secure Boot (OSB).  If not, see <http://www.gnu.org/licenses/>.

  Implements the interface functions to access the TPM
 =====================================================================================================================*/

/* ================ Global include files =============================================================================*/

#include "bios.h"
#include "tpm_driver.h"
#include "shared.h"
#include "tss_error.h"
#include "tpm_error_messages.h"

/* ================ Local constant definitions =======================================================================*/
#define PCR_MAP_SIZE                    (3U)   // Use 3 bytes (24-bits) for the PCR map size
#define NR_TPM_LOCALITIES               (5U)

// access bits
#define ACCESS_ACTIVE_LOCALITY          (0x20) // (R)
#define ACCESS_RELINQUISH_LOCALITY      (0x20) // (W)
#define ACCESS_REQUESt_USE              (0x02) // (W)

// status bits
#define STS_VALID                       (0x80) // (R)
#define STS_COMMAND_READY               (0x40) // (R)
#define STS_DATA_AVAIL                  (0x10) // (R)
#define STS_DATA_EXPECT                 (0x08) // (R)
#define STS_GO                          (0x20) // (W)

// offsets of send/receive data
#define CMD_HEADER_SIZE                 (6U)
#define CMD_RESULT_OFFSET               (6U)
#define CMD_PCR_OFFSET                  (10U)
#define CMD_REC_LEN_OFFSET              (2U)
#define CMD_SEND_LEN_OFFSET             (2U)

// receive command buffer size
#define REC_BUF_SIZE                    (4096)

// probability of performing additional PCR read (security improvement)
#define TPM_PCR_READ_PROB               (5)

// length of the return code string
#define RET_CODE_STR_LEN                (5)

/* ================ Local macro definitions ==========================================================================*/

// macros to access registers at locality
#define ACCESS(l)                       (0x0000U | (((t_U32)l) << 12U))
#define STS(l)                          (0x0018U | (((t_U32)l) << 12U))
#define DATA_FIFO(l)                    (0x0024U | (((t_U32)l) << 12U))
#define DID_VID(l)                      (0x0F00U | (((t_U32)l) << 12U))

// swap macros
#define SWAP32(x)                       (((((t_U32)x) & 0xffU) << 24U) | ((((t_U32)x) & 0xff00U) << 8U) | ((((t_U32)x) & 0xff0000U) >> 8U) | ((((t_U32)x) >> 24U) & 0xffU))
#define SWAP16(x)                       ((((((t_U16)x) & 0x00ffU) << 8U) | ((((t_U16)x) & 0xff00U) >> 8U)))

// memory area of tpm
#define TPM_ADDR                        ((t_U32)0xfed40000)

// timeout of tpm response
#define TPM_TIMEOUT_MS                  ((t_U32)5)

/* ================ Local type definitions ===========================================================================*/
typedef struct s_tpmHdr {
    t_U16 cmd;
    t_U32 size;
    t_U32 subCmd;
} PACKED_BYTE_ALIGN(T_TPM_CMD_HEADER);

/* ================ Local data definitions ===========================================================================*/

static t_U8 gLocality = 0;
static t_U8 recBuf[REC_BUF_SIZE];

/* ================ Local functions ==================================================================================*/

static INLINE t_U8 tpm_read8(t_U32 const addr)
{
    return *(t_U8 *) ((t_U32) TPM_ADDR | addr);
}
static INLINE void tpm_write8(t_U8 const val, t_U32 const addr)
{
    *(t_U8 *) ((t_U32) TPM_ADDR | addr) = val;
}

static INLINE t_U16 tpm_read16(t_U32 const addr)
{
    return *(t_U16 *) ((t_U32) TPM_ADDR | addr);
}

static INLINE t_U32 tpm_read32(t_U32 const addr)
{
    return *(t_U32 *) ((t_U32) TPM_ADDR | addr);
}

static t_U32 tpm_getReturnCode(void)
{
    t_U32 receivedReturnCodeRaw;
    memcpy(&receivedReturnCodeRaw, recBuf + CMD_RESULT_OFFSET, sizeof(receivedReturnCodeRaw));
    return SWAP32(receivedReturnCodeRaw);
}

static int tpm_checkReturnCode(t_U32 const returnCode)
{
    t_U32 const receivedReturnCode = tpm_getReturnCode();
#ifdef DEBUG
    printf("tpm_checkReturnCode: %d\n", receivedReturnCode);
#endif
    if (receivedReturnCode != returnCode) {
        char receivedReturnCodeStr[RET_CODE_STR_LEN] = {0};
        sprintf(receivedReturnCodeStr, "Code 0x%l %s", receivedReturnCode, tpm_getErrorDescription(receivedReturnCode));
        UPDATE_LOG_WARN_EXT(ERR_TPM_RETURN_CODE, receivedReturnCodeStr);
        return OMN_EXIT_FAILURE;
    }
    return OMN_EXIT_SUCCESS;
}

static int tpm_requestLocality(t_U8 const l)
{
    tpm_write8(ACCESS_RELINQUISH_LOCALITY, ACCESS(gLocality));
    tpm_write8(ACCESS_REQUESt_USE, ACCESS(l));
    if (tpm_read8(ACCESS(l) & ACCESS_ACTIVE_LOCALITY)) {
        gLocality = l;
        return gLocality;
    }
    return OMN_EXIT_FAILURE;
}

static TSS_BOOL tpm_checkStatusCondition(t_U16 const mask, t_U16 const condition)
{
    t_U16 const status = tpm_read16(STS(gLocality));
    if ((status & mask) != condition) {
        return TRUE;
    }
    return FALSE;
}

static int tpm_waitStatusCondition(t_U16 const mask, t_U16 const condition)
{
    int status = OMN_EXIT_SUCCESS;
    TSS_BOOL conditionValid;
	bios_timerInit(TPM_TIMEOUT_MS);
    do {
		int timerOverflow = bios_timerOverflow();
        conditionValid = tpm_checkStatusCondition(mask, condition);
        if (timerOverflow == TRUE) {
            UPDATE_LOG_WARN(ERR_TPM_TIMEOUT);
            status = OMN_EXIT_FAILURE;
        }
    } while ((ISFALSE(conditionValid)) && (status != OMN_EXIT_FAILURE));

    return status;
}

/**
 * Send TPM command
 * @param buf [in] data to send
 * @param bufLen [in] length of data
 * @return successful completion of operation
 */
static int tpm_send(t_U8 const* const buf, t_U32 const bufLen)
{
    int burstcnt = 0;
    int count = 0;
    TSS_BOOL conditionValid;

    // if we cannot obtain a locality
    int status = tpm_requestLocality(gLocality);
    if (status == OMN_EXIT_SUCCESS) {

        // inform we are ready to transmit data
        tpm_write8(STS_COMMAND_READY, STS(gLocality));

        // wait for TPM to become ready
        status = tpm_waitStatusCondition(STS_COMMAND_READY, 0);
        if (status == OMN_EXIT_SUCCESS) {

            while (count < bufLen - 1) {
                burstcnt = tpm_read8(STS(gLocality) + 1U);
                burstcnt += (t_U32) tpm_read8(STS(gLocality) + 2U) << 8U;
                if (burstcnt != 0) {
                    while ((burstcnt > 0) && (count < bufLen - 1)) {
                        tpm_write8(buf[count], DATA_FIFO(gLocality));
                        count++;
                    }

                    // check for overflow
                    status = tpm_waitStatusCondition(STS_VALID, 0);
                    if (status == OMN_EXIT_FAILURE) {
                        break;
                    }

                    // if there is no data expected
                    conditionValid = tpm_checkStatusCondition(STS_DATA_EXPECT, 0);
                    if (ISFALSE(conditionValid)) {
                        status = OMN_EXIT_FAILURE;
                        break;
                    }

                }
            }
            tpm_write8(buf[count], DATA_FIFO(gLocality));

            // check for overflow
            status = tpm_waitStatusCondition(STS_VALID, 0);
            if (status == OMN_EXIT_SUCCESS) {
                // if there is data expected
                conditionValid = tpm_checkStatusCondition(STS_DATA_EXPECT, 0);
                if (ISTRUE(conditionValid)) {
                    status = OMN_EXIT_FAILURE;
                } else {
                    tpm_write8(STS_GO, STS(gLocality));
                }
            }
        }
    }
    return status;
}

/**
 * receive data from TPM
 * @param buf [out] received data
 * @param bufLen [in] length of buffer
 * @param expected [in/out] In: expected number of bytes we are going to receive. Out: size of received data
 * @return successful completion of operation
 */
static int tpm_recvData(t_U8* const buf, t_U32 const bufLen, t_U32* const expected)
{
    int status = OMN_EXIT_SUCCESS;
    t_U32 size = 0, burstcnt = 0;
    TSS_BOOL conditionValid = tpm_checkStatusCondition(STS_DATA_AVAIL | STS_VALID, 0);
    while ((ISTRUE(conditionValid)) && (size < *expected)) {
        if (burstcnt == 0) {
            burstcnt = (t_U32) tpm_read8(STS(gLocality) + 1U);
            burstcnt += (t_U32) tpm_read8(STS(gLocality) + 2U) << 8U;
        }
        while ((burstcnt > 0) && (size < *expected)) {
            if (size < bufLen) {
                buf[size] = tpm_read8(DATA_FIFO(gLocality));
                size++;
            } else {
                UPDATE_LOG_ERR(ERR_BUFFER_OVERFLOW);
                status = OMN_EXIT_FAILURE;
                break;
            }
        }
    }
    *expected = size;

    return status;
}

/**
 * receive command from TPM
 * @param buf [out] received data
 * @param bufLen [in] length of buffer
 * @return successful completion of operation
 */
static int tpm_recv(t_U8* const buf, t_U32 const bufLen)
{

    t_U32 size = CMD_HEADER_SIZE, sizeTotal = 0;

    // if there is no more data available
    int status = tpm_waitStatusCondition(STS_DATA_AVAIL, 0);
    if (status == OMN_EXIT_SUCCESS) {
        // if there is no more data available or valid
        TSS_BOOL conditionValid = tpm_checkStatusCondition(STS_DATA_AVAIL | STS_VALID, 0);
        if (ISTRUE(conditionValid)) {
            // if the received data is smaller then the expected header size
            status = tpm_recvData(buf, bufLen, &size);
            sizeTotal = size;
            if ((size < CMD_HEADER_SIZE) || (status == OMN_EXIT_FAILURE)) {
                status = OMN_EXIT_FAILURE;
            } else {
                // get the expected number of bytes that we are going to receive
                t_U32 expected;
                (void)memcpy(&expected, buf + CMD_REC_LEN_OFFSET, sizeof(expected));
                expected = SWAP32(expected);
                if (expected >= bufLen) {
                    status = OMN_EXIT_FAILURE;
                } else {
                    // if the total received bytes is smaller then the remaining we expect
                    size = expected - CMD_HEADER_SIZE - 1;
                    status = tpm_recvData(recBuf + CMD_HEADER_SIZE, bufLen - CMD_HEADER_SIZE, &size);
                    sizeTotal += size;
                    if ((sizeTotal < (expected) - 1) || (status == OMN_EXIT_FAILURE)) {
                        status = OMN_EXIT_FAILURE;
                    } else {
                        // if there is no more data available or valid
                        conditionValid = tpm_checkStatusCondition(STS_DATA_AVAIL | STS_VALID, 0);
                        if (ISFALSE(conditionValid)) {
                            status = OMN_EXIT_FAILURE;
                        } else {
                            // if the total received bytes is smaller then the remaining we expect
                            size = 1;
                            status = tpm_recvData(buf + sizeTotal, bufLen - sizeTotal, &size);
                            sizeTotal += size;
                            if ((sizeTotal != expected) && (status == OMN_EXIT_FAILURE)) {
                                status = OMN_EXIT_FAILURE;
                            } else {
                                // if there is no more data available or valid
                                conditionValid = tpm_checkStatusCondition(STS_DATA_AVAIL | STS_VALID, 0);
                                if (ISFALSE(conditionValid)) {
                                    status = OMN_EXIT_FAILURE;
                                } else {
                                    // inform we are ready
                                    tpm_write8(STS_COMMAND_READY, STS(gLocality));
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    return status;
}

/**
 * send and receive TPM command
 * @param sendBuf [out] received data
 * @param sendBufLen [in] length of buffer
 * @return successful completion of operation
 */
static int tpm_commandWriteRead(void *sendBuf, t_U32 sendBufLen)
{
    t_U32 swappedSendBufLen = SWAP32(sendBufLen);
    memcpy((t_U8*)sendBuf + CMD_SEND_LEN_OFFSET, &swappedSendBufLen, sizeof(swappedSendBufLen));
    int status = tpm_send((t_U8*)sendBuf, sendBufLen);
    if (status == OMN_EXIT_SUCCESS) {
        status = tpm_recv(recBuf, REC_BUF_SIZE);
        if (status != OMN_EXIT_SUCCESS) {
            t_U32 const returnCode = tpm_getReturnCode();
            tpm_checkReturnCode(returnCode);
        }
    }
    return status;
}

/* ================ Global functions (Global system scope) ===========================================================*/
/**
 * get TPM vendor
 * @return vendor identifier of TPM (see  TPM_VEN_* defines)
 */
t_U32 tpm_getVendorId(void)
{
    return tpm_read32(DID_VID(0)) & 0xFFFFU;
}

/**
 * Initialize TPM
 * @return exit state indicating if initialization was successful
 */
int tpm_init(void)
{
    int status = OMN_EXIT_SUCCESS;
    t_U32 vendor;
    unsigned int i;

    for (i = 0; i < NR_TPM_LOCALITIES; i++) {
        tpm_write8(ACCESS_RELINQUISH_LOCALITY, ACCESS(i));
    }

    if (tpm_requestLocality(0) < 0) {
        return OMN_EXIT_FAILURE;
    }

    vendor = tpm_getVendorId();
    if (vendor == TPM_VEN_INFINEON) {
#ifdef DEBUG
        printf("Detected Infineon TPM (0x%X)\n", vendor);
#endif
    } else {
#ifdef DEBUG
        printf("Detected unspecified TPM (0x%X)\n", vendor);
#endif
    }

    return status;
}

/**
 * Perform TPM self test
 * @return exit state indicating if self-test was successful
 */
int tpm_selftest(void)
{
    struct {
        T_TPM_CMD_HEADER hdr;
    } PACKED_BYTE_ALIGN(tpmCmd);

    tpmCmd.hdr.cmd    = SWAP16(TPM_TAG_RQU_COMMAND);
    tpmCmd.hdr.size   = SWAP32((t_U32)sizeof(tpmCmd));
    tpmCmd.hdr.subCmd = SWAP32(TPM_ORD_SelfTestFull);

    tpm_commandWriteRead(&tpmCmd, sizeof(tpmCmd));

    return OMN_EXIT_SUCCESS;
}

/**
 * Start TPM
 * @param startupType [in] startup type of TPM (see TPM_ST_* defines)
 * @return exit state indicating if startup was successful
 */
int tpm_startup(t_U16 const startupType)
{
    int status;

    struct {
        T_TPM_CMD_HEADER hdr;
        t_U16            startupType;
    } PACKED_BYTE_ALIGN(tpmCmd);

    tpmCmd.hdr.cmd    = SWAP16(TPM_TAG_RQU_COMMAND);
    tpmCmd.hdr.size   = SWAP32((t_U32)sizeof(tpmCmd));
    tpmCmd.hdr.subCmd = SWAP32(TPM_ORD_Startup);
    tpmCmd.startupType= SWAP16(startupType);

    status = tpm_commandWriteRead(&tpmCmd, sizeof(tpmCmd));

    if (status == OMN_EXIT_SUCCESS) {
        status = tpm_checkReturnCode(TSS_SUCCESS);
    }

    return status;
}

/**
 * Read PCR value from TPM
 * @param index [in] index of PCR
 * @param data  [out] data of PCR (at least of size PCR_SIZE)
 * @return exit state indicating if reading of PCR was successful
 */
int tpm_readPcr(t_U32 const index, t_U8* const data)
{
    int status;

    struct {
        T_TPM_CMD_HEADER hdr;
        t_U32            pcrIndex;
    } PACKED_BYTE_ALIGN(tpmCmd);

    tpmCmd.hdr.cmd    = SWAP16(TPM_TAG_RQU_COMMAND);
    tpmCmd.hdr.size   = SWAP32((t_U32)sizeof(tpmCmd));
    tpmCmd.hdr.subCmd = SWAP32(TPM_ORD_PcrRead);
    tpmCmd.pcrIndex   = SWAP32(index);

    status = tpm_commandWriteRead(&tpmCmd, sizeof(tpmCmd));
    if (status == OMN_EXIT_SUCCESS) {
        status = tpm_checkReturnCode(TSS_SUCCESS);
        if (status == OMN_EXIT_SUCCESS) {
            memcpy(data, recBuf + CMD_PCR_OFFSET, PCR_SIZE);
        }
    }
    return status;
}

/**
 * Read PCR value from TPM and confuse attacker by performing additional reads that are not used
 * @param index [in] index of PCR
 * @param data  [out] data of PCR (at least of size PCR_SIZE)
 * @return exit state indicating if reading of PCR was successful
 */
int tpm_readPcrObscure(t_U32 const index, t_U8* const data) {
	int status;
	t_U32 indexDummy;
	t_U8 dummy[PCR_SIZE];

	// fake read
	while (utils_pseudoRand() % TPM_PCR_READ_PROB != 0) {
		indexDummy = utils_pseudoRand() % (t_U32)TPM_PCRS;
		(void) tpm_readPcr(indexDummy, dummy);
	}

	// real read
	status = tpm_readPcr(index, data);

	// fake read
	while (utils_pseudoRand() % TPM_PCR_READ_PROB != 0) {
		indexDummy = utils_pseudoRand() % (t_U32)TPM_PCRS;
		(void) tpm_readPcr(indexDummy, dummy);
	}
	return status;
}

/**
 * Hash PCR with data
 * @param index         [in] PCR index
 * @param data          [in] data to hash PCR with
 * @param dataSize      [in] size of data
 * @return exit state indicating if hashing was successful
 */
int tpm_extendPcr(t_U8 const index, t_U8 const * const data, t_U32 const dataSize)
{
    int status;

    struct {
        T_TPM_CMD_HEADER hdr;
        t_U32            pcrIndex;
        t_U8             data[dataSize];
    } PACKED_BYTE_ALIGN(tpmCmd);

    tpmCmd.hdr.cmd    = SWAP16(TPM_TAG_RQU_COMMAND);
    tpmCmd.hdr.size   = SWAP32((t_U32)sizeof(tpmCmd));
    tpmCmd.hdr.subCmd = SWAP32(TPM_ORD_Extend);
    tpmCmd.pcrIndex   = SWAP32(index);
    (void)memcpy(tpmCmd.data, data, sizeof(tpmCmd.data));

    status = tpm_commandWriteRead(&tpmCmd, sizeof(tpmCmd));
    if (status == OMN_EXIT_SUCCESS) {
        status = tpm_checkReturnCode(TSS_SUCCESS);
    }
    return status;
}

/**
 * Reset PCR
 * @param index           [in] PCR index to reset
 * @return exit state indicating if reset was successful
 */
int tpm_resetPcr(t_U8 const index)
{
    int status = OMN_EXIT_FAILURE;

    if ((index / 8U) < PCR_MAP_SIZE) {
        struct {
            T_TPM_CMD_HEADER hdr;
            t_U16            mapSize;
            t_U8             map[PCR_MAP_SIZE];
        } PACKED_BYTE_ALIGN(tpmCmd);

        tpmCmd.hdr.cmd    = SWAP16(TPM_TAG_RQU_COMMAND);
        tpmCmd.hdr.size   = SWAP32((t_U32)sizeof(tpmCmd));
        tpmCmd.hdr.subCmd = SWAP32(TPM_ORD_PCR_Reset);
        tpmCmd.mapSize    = SWAP16((t_U16)sizeof(tpmCmd.map));
        // set bitmap with correct PCR index
        tpmCmd.map[index / 8U] = 0x01U < (index % 8U);

        status = tpm_commandWriteRead(&tpmCmd, sizeof(tpmCmd));
        if (status == OMN_EXIT_SUCCESS) {
            status = tpm_checkReturnCode(TSS_SUCCESS);
        }
    }

    return status;
}

/**
 * Read from TPM nvmem area
 * @param area          [in] location of the nvmem area
 * @param offset        [in] offset of data
 * @param data          [in] data to read
 * @param dataSize      [in] size of data to read
 * @return exit state indicating if read was successful
 */
int tpm_nvread(t_U32 const area, t_U32 const offset, t_U8 * const data, t_U32 const dataSize)
{
    int status;

    struct {
        T_TPM_CMD_HEADER hdr;
        t_U32            area;
        t_U32            offset;
        t_U32            dataSize;
    } PACKED_BYTE_ALIGN(tpmCmd);

    tpmCmd.hdr.cmd    = SWAP16(TPM_TAG_RQU_COMMAND);
    tpmCmd.hdr.size   = SWAP32((t_U32)sizeof(tpmCmd));
    tpmCmd.hdr.subCmd = SWAP32(TPM_ORD_NV_ReadValue);
    tpmCmd.area       = SWAP32(area);
    tpmCmd.offset     = SWAP32(offset);
    tpmCmd.dataSize   = SWAP32(dataSize);

    status = tpm_commandWriteRead(&tpmCmd, sizeof(tpmCmd));

    if (status == OMN_EXIT_SUCCESS) {
        status = tpm_checkReturnCode(TSS_SUCCESS);

        if (status == OMN_EXIT_SUCCESS) {
            memcpy(data, recBuf + 14, (int) dataSize);
        }
    }

    return status;
}

/**
 * Write to TPM nvmem area
 * @param area          [in] location of the nvmem area
 * @param offset        [in] offset of data
 * @param data          [in] data to write
 * @param dataSize      [in] size of the nvmem area
 * @return exit state indicating if write was successful
 */
int tpm_nvwrite(t_U32 const area, t_U32 const offset, t_U8 const* const data, t_U32 const dataSize)
{
    int status;

    struct {
        T_TPM_CMD_HEADER hdr;
        t_U32            area;
        t_U32            offset;
        t_U32            dataSize;
        t_U8             data[dataSize];
    } PACKED_BYTE_ALIGN(tpmCmd);

    tpmCmd.hdr.cmd    = SWAP16(TPM_TAG_RQU_COMMAND);
    tpmCmd.hdr.size   = SWAP32((t_U32)sizeof(tpmCmd));
    tpmCmd.hdr.subCmd = SWAP32(TPM_ORD_NV_WriteValue);
    tpmCmd.area       = SWAP32(area);
    tpmCmd.offset     = SWAP32(offset);
    tpmCmd.dataSize   = SWAP32(dataSize);
    (void)memcpy(tpmCmd.data, data, sizeof(tpmCmd.data));

    status = tpm_commandWriteRead(&tpmCmd, sizeof(tpmCmd));

    if (status == OMN_EXIT_SUCCESS) {
        status = tpm_checkReturnCode(TSS_SUCCESS);
    }
    return status;
}

/* ===================================================================================================================*/

/* ================ End of file ======================================================================================*/
