/* =====================================================================================================================
 (c) Copyright OMRON Europe BV 2016
 All Right Reserved
 OMRON Development & Engineering
 ================================================================================================================= *//**
 @file

  This file is part of Omron Secure Boot (OSB).

  Omron Secure Boot (OSB) is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Omron Secure Boot (OSB) is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Omron Secure Boot (OSB).  If not, see <http://www.gnu.org/licenses/>.

 Contains supporting functions for other components

 @internal
 *//* ================================================================================================================ */
#include "utils.h"
#include "shared.h"

/* ================ Local constant definitions =======================================================================*/

#define RND_MUL         (279470273UL)
#define RND_MOD         (4294967291UL)

/* ================ Local macro definitions ==========================================================================*/

/* ================ Local type definitions ===========================================================================*/

/* ================ Local data definitions ===========================================================================*/

/**
 * Initialize random number with prime
 */
static t_U32 rndValue = 5;

/* ================ Local functions ==================================================================================*/

/**
 * Converts @p c to upper-case
 * @param c [in] The lower-case character
 * @return the upper-case character
 */
static char toUpper(char const c) {
    if (('a' <= c) && (c <= 'z')) {
        return 'A' + (c - 'a');
    }
    return c;
}

/**
 * Will return @p TRUE when @p 'c' is a hex-digit
 * @param c [in] The character to test
 * @return @p TRUE when it contains a hex-digit
 */
static t_bool isHex(char const c) {
    char const cUpper = toUpper(c);

    return (t_bool) (((cUpper >= 'A' && cUpper <= 'F') || (cUpper >= '0' && cUpper <= '9')) ? TRUE : FALSE);

}

/**
 * Converts a hex-digit to a byte
 * @param c [in] The hex-digit
 * @return the byte value
 */
static t_U8 hexDigitToByte(char const c) {
    char const cUpper = (t_U8) toUpper(c);
    t_U8 byte = 0;

    if ((cUpper >= 'A' && cUpper <= 'F')) {
        byte = ((t_U8) (cUpper - 'A')) + 0x0AU;
    } else if ((cUpper >= '0' && cUpper <= '9')) {
        byte = ((t_U8) (cUpper - '0'));
    }
    return byte;
}

/**
 * Converts a byte to a hex-digit
 * @param b [in] The byte to convert
 * @return The hex-digit
 */
static char bytetoHexDigit(t_U8 const b) {
    char c = ' ';

    if ((b >= 0) && (b <= 9)) {
        c = '0' + b;
    }
    if ((b >= 0x0AU) && (b <= 0x0FU)) {
        c = 'A' + (b - 0x0AU);
    }
    return c;
}

/* ================ Global functions (Global system scope) ===========================================================*/

/**
 * Efficient buffer replace for large buffers  [dataStart][skipBytes][dataEnd] => [dataStart][bufRep][dataEnd]
 * @param buf [in] buffer of size bufsize [out] buffer with replaced string
 * @param bufsize [in] size of buffer in bytes
 * @param skipBytes [in] number of bytes to skip
 * @param bufRep [in] replacement string
 * @param bufRepSize [in] size of replacement buffer
 * @return exit status
 */
int utils_bufReplace(char* const pos, int const bufsize, int const skipBytes, char const* const bufRep, int const bufRepSize) {
    // calculate remainder of part that we are going to move
    int remainder = strlen(pos + skipBytes);
    // there should be enough space left in the buffer if we need to replace it with a larger string
    if (remainder >= bufsize - bufRepSize) {
        UPDATE_LOG_WARN(ERR_WONT_FIT);
        return OMN_EXIT_FAILURE;
    }
    // make room for string that we are going to replace with
    memcpy(pos + bufRepSize, pos + skipBytes, remainder);
    // write replacement string
    memcpy(pos, bufRep, bufRepSize);
    return OMN_EXIT_SUCCESS;
}

/**
 * Efficient string replace for large buffers
 * @param buf [in] \0 terminating buffer of size bufsize [out] buffer with replaced strings
 * @param bufsize [in] size of buffer in bytes
 * @param strFind [in] string to search for
 * @param strRep [in] replacement string
 * @return void
 */
void utils_strReplace(char *buf, int const bufsize, char const* const strFind, char const* const strRep) {
    int i;
    // iterate through end of string
    for (i = 0; i < strlen(buf); i++) {
        // store current position
        char* const pos = buf + i;
        if (memcmp(pos, strFind, strlen(strFind)) == 0) {
            utils_bufReplace(pos, bufsize-i, strlen(strFind), strRep, strlen(strRep));
            // move position
            i += strlen(strRep);
        }
    }
}


/**
 * Read hex string as single byte
 * @param hexStr [in]  hexStr to read from (hexStr should be a allocated character pointer of at least 2 bytes)
 * @param byte   [out] byte read from hexStr
 * @return void
 */
void utils_hexStrToByte(char const * const hexStr, t_U8* const byte) {
    *byte = (hexDigitToByte(hexStr[0]) << 4) + hexDigitToByte(hexStr[1]);
}

/**
 * Print single byte as hex string
 * @param byte   [in] byte to be printed
 * @param hexStr [out] byte printed in hexStr (hexStr should be a allocated character pointer of at least 2 bytes)
 * @return void
 */
void utils_byteToHexStr(t_U8 const byte, char* const hexStr) {
    t_U8 const high = (byte & 0xF0U) >> 4;
    t_U8 const low = (byte & 0x0FU);

    hexStr[0] = bytetoHexDigit(high);
    hexStr[1] = bytetoHexDigit(low);
}

/**
 * Prints the data in hex-text (lowest byte at the beginning of the string)
 * @param data [in] The data to print
 * @param len  [in] The length in bytes of data
 * @return void
 */
void utils_printAsHexStr(t_U8 const* const data, int const len) {
    int i;
    for (i = 0; i < len; i++) {
        char hexStr[3] = { 0 };
        utils_byteToHexStr(data[i], hexStr);
        printf(hexStr);
    }
}

/**
 * Prints key in LSB
 * @param key    [in]   Key to print
 * @param keyLen [in]   Length of key
 * @return void
 */
void utils_printKeyAsHexStr(t_U8 const* const key, int const keyLen) {
    int i;
    for (i = keyLen - 1; i >= 0; i--) {
        char hexStr[3] = { 0 };
        utils_byteToHexStr(key[i], hexStr);
        printf(hexStr);
    }
}

/**
 * Prints the data in reverse hex-text (lowest byte at the end of the string)
 * @param data [in] The data to print
 * @param len  [in] The length in bytes of data
 * @return void
 */
void utils_printAsReverseHexStr(t_U8 const* const data, int const len) {
    int i;
    for (i = len - 1; i >= 0; i--) {
        char hexStr[2] = { 0 };
        utils_byteToHexStr(data[i], hexStr);
        putchar(hexStr[0]);
        putchar(hexStr[1]);
    }
}

/**
 * Converts a hex-string to a byte array
 * @param data      [out] The converted data from the @p hexstring
 * @param dataLen   [inout]  IN: The length of the data-buffer; OUT: The actual data written
 * @param hexstring [in]  The string containing hex values
 * @return exit state indicating if the hexstring conversions was successful
 */
int utils_hexStrToArrayReverse(t_U8* const data, t_U32* const dataLen, char const * const hexstring)
{
    t_bool   haveError = FALSE;
    unsigned i         = 0;

    if ((*dataLen > 0) && ((strlen(hexstring) % 2) == 0)) {
        while (ISFALSE(haveError) && (hexstring[i*2] != '\0') && (hexstring[i*2] != ' ') && (i < *dataLen)) {
            t_bool const isHexDigitH = isHex(*(char const * const ) (hexstring + i * 2));
            t_bool const isHexDigitL = isHex(*(char const * const ) (hexstring + i * 2 + 1));

            if (ISTRUE(isHexDigitH) && ISTRUE(isHexDigitL)) {
                utils_hexStrToByte(hexstring + i * 2, &data[*dataLen - 1 - i]);
                i++;
            } else {
                haveError = TRUE;
            }
        }
        *dataLen = i;

        if ((hexstring[i*2] != ' ') && (hexstring[i*2] != '\0')) {
            //Not the full hex-string is converted
            haveError = TRUE;
        }
    } else {
        haveError = TRUE;
    }

    return (int)(ISTRUE(haveError) ? OMN_EXIT_FAILURE : OMN_EXIT_SUCCESS);
}

/**
 * Converts a hex-string to a byte array
 * @param data      [out] The converted data from the @p hexstring
 * @param dataLen   [inout]  IN: The length of the data-buffer; OUT: The actual data written
 * @param hexstring [in]  The string containing hex values
 * @return exit state indicating if the hexstring conversions was successful
 */
int utils_hexStrToArray(t_U8* const data, t_U32* const dataLen, char const * const hexstring)
{
    t_bool   haveError = FALSE;
    unsigned i         = 0;

    if ((*dataLen > 0) && ((strlen(hexstring) % 2) == 0)) {
        while (ISFALSE(haveError) && (hexstring[i*2] != '\0') && (hexstring[i*2] != ' ') && (i < *dataLen)) {
            t_bool const isHexDigitH = isHex(*(char const * const ) (hexstring + i * 2));
            t_bool const isHexDigitL = isHex(*(char const * const ) (hexstring + i * 2 + 1));

            if (ISTRUE(isHexDigitH) && ISTRUE(isHexDigitL)) {
                utils_hexStrToByte(hexstring + i * 2, &data[i]);
                i++;
            } else {
                haveError = TRUE;
            }
        }
        *dataLen = i;

        if ((hexstring[i*2] != ' ') && (hexstring[i*2] != '\0')) {
            //Not the full hex-string is converted
            haveError = TRUE;
        }
    } else {
        haveError = TRUE;
    }

    return (int)(ISTRUE(haveError) ? OMN_EXIT_FAILURE : OMN_EXIT_SUCCESS);
}

/**
 * Pseudo random number generator based on prime values
 * @return returns the random number
 */
t_U32 const utils_pseudoRand(void) {
    rndValue = (rndValue * RND_MUL) % RND_MOD;
    return rndValue;
}

/**
 * Converts a numerical string to an integer
 * @param str [in] String (can contain a negative value)
 * @return The numerical integer
 */
int util_atoi(char const * const str)
{
    int res = 0;
    int sign = 1;
    int i = 0;

    if (str[0] == '-') {
        sign = -1;
        i++;
    }

    for (; ((str[i] != '\0') && (str[i] != ' ')); i++) {
        res = res*10 + str[i] - '0';
    }

    // Return result with sign
    return sign*res;
}
/* ===================================================================================================================*/

/* ================ End of file ======================================================================================*/

