/* =====================================================================================================================
(c) Copyright OMRON Europe BV 2016
All Right Reserved 
OMRON Development & Engineering 
================================================================================================================== *//**
@file 

This file is part of Omron Secure Boot (OSB).

 Omron Secure Boot (OSB) is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Omron Secure Boot (OSB) is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Omron Secure Boot (OSB).  If not, see <http://www.gnu.org/licenses/>.

  Implements higher level functions to communicate with a serial device

@internal
*//* ================================================================================================================ */ 

/* ================ Global include files ============================================================================ */

/* ================ Sub-system include files ======================================================================== */
#include "comm.h"
#include "serial.h"

/* ================ Local constant definitions ====================================================================== */
#define SYNC_TIMEOUT_CNT         (100000)

#define SYNC_BYTE                (0x00U)
#define SYNC_RSP_BYTE            (0xACU)

/* ================ Local macro definitions ========================================================================= */

/* ================ Local type definitions ========================================================================== */

/* ================ Local data definitions ========================================================================== */

/* ================ Local functions ================================================================================= */

/* ================ Global functions (Sub-system scope) ============================================================= */

/* ================ Global functions (Global system scope) ======================================================= *//**
@endinternal
*//* ================================================================================================================ */
#ifdef SUPPORT_SERIAL
/**
 * Sends arbitrary data via the serial hw port
 * @param buf       [in] The data to be transmitted
 * @param txBufLen  [in] the length of the data to be transmitted
 * @return @p OMN_EXIT_SUCCESS when the data is sent; @p OMN_EXIT_FAILURE otherwise
 */
int comm_send(void const * const buf, int const txBufLen)
{
    int                 result   = OMN_EXIT_SUCCESS;
    t_U8 const * const  charBuf  = buf;
    t_bool const        commInit = serial_is_init();


    if (ISTRUE(commInit)) {
        int i = 0;
#ifdef DEBUG
        grub_printf("Sending data. Length %d\n", txBufLen);
#endif
        comm_flush();

        while ((i < txBufLen) && (result == OMN_EXIT_SUCCESS)) {
            result = serial_hw_put(charBuf[i]);
            if (result == OMN_EXIT_SUCCESS) {
                i++;
            }
        }
    } else {
        result = OMN_EXIT_FAILURE;
    }

    return result;
}

/**
 * Receives data from the serial hw port
 * @param rxBuf      [in] The buffer to store the received data in
 * @param rxBufLen   [in] The number of expected bytes to receive.
 * @param timeoutCnt [in] A counter value which will decrement until 0. When 0, the function will give up on waiting for an answer
 * @note This function expects to receive @p rxBufLen. It return an error if not all expected bytes are received.
 * @return @p OMN_EXIT_SUCCESS when the data is received; @p OMN_EXIT_FAILURE otherwise
 */
int comm_receive(void* const rxBuf, int const rxBufLen, int const timeoutCnt)
{
    int             result    = OMN_EXIT_FAILURE;
    t_U8* const     charRxBuf = rxBuf;
    t_bool const    commInit  = serial_is_init();

    if (ISTRUE(commInit)) {
        int i = 0;
        int retry = timeoutCnt; //TODO This 'timeout' can be implemented better

        while ((retry > 0) && (i < rxBufLen)) {
            int c = serial_hw_fetch();

            if (c >= 0) {
                charRxBuf[i] = c;
                i++;
            } else {
                retry--;
            }
        }

        if (i >= rxBufLen) {
            result = OMN_EXIT_SUCCESS;
        } else {
#ifdef DEBUG
            grub_printf("RX error; could not receive all bytes. Receive count: %d from %d\n", i, rxBufLen);
#endif
        }
    }
    return result;
}

/**
 * Flushes the RX buffer
 */
void comm_flush(void)
{
	int c = 0;

	while (c >= 0) {
		c = serial_hw_fetch();
	}
}

/**
 * Synchronizes the communication device attached to the serial hw port
 * @param retryCnt [in] The number of sync retries before giving up
 * @return @p TRUE when the device is synchornized; @p FALSE otherwise
 */
t_bool comm_sync(t_U16 const retryCnt)
{
    t_bool       isSynced = FALSE;
    t_bool const commInit = serial_is_init();

    if (ISTRUE(commInit)) {
        t_U8 const txBuf[] = { SYNC_BYTE };
        t_U8       rxResult;
        int        result;
        int        txResult;
        int        retryCount = (int)retryCnt;

        comm_flush();
        do {
            txResult = comm_send(txBuf, sizeof(txBuf));
            if (txResult == OMN_EXIT_SUCCESS) {
                result = comm_receive(&rxResult, sizeof(rxResult), SYNC_TIMEOUT_CNT);
                //grub_printf("comm_receive status: %d; received data: %x\n", result, rxResult);
                if (result == OMN_EXIT_SUCCESS) {
                    if (rxResult != SYNC_RSP_BYTE) {
                        result = OMN_EXIT_FAILURE;
                    }
                }
                retryCount--;
            }
        } while ((txResult == OMN_EXIT_SUCCESS) && (result != OMN_EXIT_SUCCESS) && (retryCount > 0));

        if ((retryCount > 0) && (txResult == OMN_EXIT_SUCCESS) && (result == OMN_EXIT_SUCCESS)) {
            isSynced = TRUE;
        }
    }

    return isSynced;
}

#endif
/* ================ End of file ===================================================================================== */
