/* =====================================================================================================================
(c) Copyright OMRON Europe BV 2016
All Right Reserved 
OMRON Development & Engineering 
================================================================================================================== *//**
@file 

 This file is part of Omron Secure Boot (OSB).

 Omron Secure Boot (OSB) is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Omron Secure Boot (OSB) is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Omron Secure Boot (OSB).  If not, see <http://www.gnu.org/licenses/>.

 Caches a file from disk into memory. The size of the file should be smaller than @p MAX_FILE_CACHE_SIZE otherwise
 an error-code is returned.

 The cache can be used to do file verification AND (on-the-fly) decryption to prevent Time-of-check-to-time-of-use
 attacks

@internal
*//* ================================================================================================================ */ 

/* ================ Global include files ============================================================================ */
#include "file_cache.h"
#include "shared.h"

/* ================ Sub-system include files ======================================================================== */

/* ================ Local constant definitions ====================================================================== */
#define TEST_VECTOR     (0x55U)
#define MEMBLK_SIZE     (5 * 1024 * 1024)

/* ================ Local macro definitions ========================================================================= */

/* ================ Local type definitions ========================================================================== */

/* ================ Local data definitions ========================================================================== */
static t_U8* cacheSection   = NULL;
static t_U32 cacheSize      = 0;
static t_S32 cachedFileSize = 0;

/* ================ Local functions ================================================================================= */
/**
 * Checks the fill cache memory section for consistency to make sure files are not corrupted when stored in the cache
 */
#ifdef DEBUG
static void cacheTest(void)
{
    t_bool   cacheError = FALSE;
    unsigned i          = 0;

    printf("Testing cache.");
    do {
        cacheSection[i] = TEST_VECTOR;
        if (cacheSection[i] != TEST_VECTOR) {
            cacheError = TRUE;
        } else {
            i++;
        }
        if ((i % MEMBLK_SIZE) == 0) {
            printf(".");
        }
    } while ((i < cacheSize) && (ISFALSE(cacheError)));

    if (ISFALSE(cacheError)) {
        printf(" passed\n");
    } else {
        printf(" failed (mem-addr: 0x%x)\n", cacheSection + i);
        cacheSection = NULL;
        cacheSize    = 0;
    }
}
#endif

/* ================ Global functions (Sub-system scope) ============================================================= */

/* ================ Global functions (Global system scope) ======================================================= *//**
@endinternal
*//* ================================================================================================================ */
/**
 * Initializes the cache by assigning a memory region that can be used for caching the files
 * @param cacheStartMemAddr [in] Start pointer of the memory address to be used for the file cache
 * @param cacheMemSize      [in] Size of the memory region
 */
void filecache_init(void* const cacheStartMemAddr, t_U32 const cacheMemSize)
{
    cacheSection = (t_U8* const) cacheStartMemAddr;
    cacheSize    = cacheMemSize;
#ifdef DEBUG
    printf("File cache area start: 0x%x with size %d\n", cacheSection, cacheSize);

    cacheTest();
#endif
}

/**
 * Initializes the cache by assigning static memory from a buffer
 * @param stackBuffer       [in] Buffer to assign
 * @param cacheMemSize      [in] Size of assigned buffer
 */
void filecache_stack_init(t_U8* const stackBuffer, t_U32 const cacheMemSize)
{
    cacheSection = (t_U8* const) stackBuffer;
    cacheSize    = cacheMemSize;
#ifdef DEBUG
    printf("File cache area start: 0x%x with size %d\n", cacheSection, cacheSize);
    cacheTest();
#endif
}

/**
 * Fills the cacheSection with the @p filename contents
 * @note: An optional 'onCacheFilled' handler can be provided to interact with the cache contents
 * @param filename      [in] Mandatory file-name to load the cache with
 * @param onCacheFilled [in] Optional callback to interact with the cache contents after filling it
 * @return '1' when successful, otherwise '0'
 */
int filecache_fill(char* const filename, T_ON_CACHE_FILLED const onCacheFilled)
{
    int result = 0;

    if ((cacheSection != NULL) && (cacheSize > 0))  {
        int const readRawPrev         = read_raw;
        int const loadShaForFilePrev  = load_sha_for_file;

        if (cachedFileSize != 0) {
#ifdef DEBUG
            grub_printf("Warning: Filling the cache with a file (cacheFileSize: %d) while the previous file was not closed properly. Emptying cache and continuing.\n", cachedFileSize);
#endif
            filecache_purge();
        }

        load_sha_for_file = 0;
        cache_file        = 0;
        read_raw          = 1;
        result = grub_open(filename);
        if (result != 0) {
            if ((filemax <= cacheSize) && (filemax > 0)) {
                errnum  = 0;
                filepos = 0;
#ifdef DEBUG
                grub_printf("Start reading from file to fill cache (cache start addr: 0x%x, file size: %d)\n", cacheSection, filemax);
#endif
                result = grub_read((char*)cacheSection, filemax);
                if (result > 0) {
                    cachedFileSize = filemax;
                    result         = 1;
                    errnum         = 0;
#ifdef DEBUG
                grub_printf("File cached (cache start addr: 0x%x, cachedFileSize: %d)\n", cacheSection, cachedFileSize);
#endif
                } else {
#ifdef DEBUG
                    grub_printf("Could not read from file '%s' to fill the file-cache (result=%d; filemax=%d; errnum=%d)\n", filename, result, filemax, errnum);
#endif

                    UPDATE_LOG_WARN(ERR_CACHE_ISSUE);
                    result = 0;
                }
            } else {
                if (filemax == 0) {
#ifdef DEBUG
                    grub_printf("File does not have any contents. Nothing to cache\n");
#endif
                    cachedFileSize = 0;
                    result         = 1;
                    errnum         = 0;
                } else {
#ifdef DEBUG
                    grub_printf("File '%s' is too large (file size: %u; cache size: %u); cannot be cached\n", filename, filemax, cacheSize);
#endif
                    UPDATE_LOG_WARN(ERR_CACHE_FILE_TOO_LARGE);
                    result = 0;
                }
            }
        } else {
#ifdef DEBUG
            grub_printf("Could not open file '%s' for caching (errnum=%d)\n", filename, errnum);
#endif

            UPDATE_LOG_WARN(ERR_CACHE_ISSUE);
            result = 0;
        }
        load_sha_for_file = loadShaForFilePrev;
        cache_file        = 1;
        read_raw          = readRawPrev;
        filepos           = 0;

        if (result != 0) {
            if (onCacheFilled != NULL) {
                result = onCacheFilled(cacheSection, (t_S32)filemax);
            }
        }
    }

    return result;
}

/**
 * Main interface function to read arbitrary data from cache
 * @param buf [in] The buffer to store the read contents in
 * @param len [in] The number of bytes to read
 * @return The number of bytes read from the cache; 0 if an error occured
 */
int filecache_read(char* buf, int len)
{
    if ((cacheSection != NULL) && (cachedFileSize > 0) && ((filepos + len) <= cachedFileSize)) {
        (void)memcpy(buf, &cacheSection[filepos], len);
        filepos += len;
        return len;
    } else {
#ifdef DEBUG
    grub_printf("Read part of file from cache (filepos: %d; len: %d; cachedFileSize: %d\n", filepos, len, cachedFileSize);
#endif
    }
    return 0;
}

/**
 * Closes the cacheSection by clearing all the contents
 */
void filecache_purge(void)
{
    (void)memset(cacheSection, 0, cachedFileSize);
    cachedFileSize = 0;
}
/* ================ End of file ===================================================================================== */
