/* =====================================================================================================================
 (c) Copyright OMRON Europe BV 2016
 All Right Reserved
 OMRON Development & Engineering
 ================================================================================================================== *//**
 @file

 This file is part of Omron Secure Boot (OSB).

 Omron Secure Boot (OSB) is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Omron Secure Boot (OSB) is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Omron Secure Boot (OSB).  If not, see <http://www.gnu.org/licenses/>.

 Implements logic to detect file encryption and decrypt a given file on-the-fly

 @internal
 *//* ================================================================================================================ */

/* ================ Global include files ============================================================================ */
#include "file_decrypt.h"
#include "app_aes.h"
#include "utils.h"
#include "shared.h"

/* ================ Sub-system include files ======================================================================== */

/* ================ Local constant definitions ====================================================================== */
#define AES_FILE_HDR_MAGIC          (0x05ECB007UL)
#define AES128_KEYSIZE              (16)
#define AES_ENC_FILE_HDR_VERSION    (0x0110U) // 1.1
#define AES_ECB_MODE                (0x0001U)
#define AES_CBC_MODE                (0x0002U)

#define DEC_BUF_SIZE                (64U)

/* ================ Local macro definitions ========================================================================= */
//#define SUPPORT_AES_ECB           //Enable to support older ECB encrypted files

/* ================ Local type definitions ========================================================================== */
typedef struct {
    t_U32 magic;
    t_U16 version;
    t_U16 aesMode;
    t_U16 aesKeySize;
    t_U32 fileSize;
}PACKED_BYTE_ALIGN(t_aesFileHdr);

/* ================ Local data definitions ========================================================================== */
static struct {
    int lastDecBlockNr;
    char inputBlock[DEC_BUF_SIZE];
    char decryptedBlock[DEC_BUF_SIZE];
    t_decryptEngine decryptEngine;
    int keyLength;
} fileDecrypt = { 0 };

/* ================ Local functions ================================================================================= */
/**
 * This functions sets the decryption engine to use for on-the-fly decryption
 * This should be set before on-the-fly decryption is possible
 * @note The decryption engine is responsible to use the correct decryption key
 * @param decryptEngine [in] A function/delegate pointer for decrypting pieces of file data
 * @param keyLength     [in] The size of the key/decryption block used by the decryption engine
 */
static void setDecryptionEngine(t_decryptEngine const decryptEngine, unsigned int const keyLength)
{
    fileDecrypt.decryptEngine = decryptEngine;
    fileDecrypt.keyLength = keyLength;

    if (keyLength > sizeof(fileDecrypt.decryptedBlock)) {
        GRUB_PANIC("Internal decryption buffer is too small for the provided decryption key size.");
    }
}

/**
 * Decypts the contents in @p cachedFile and puts the result in @p cachedFile
 * @param cachedFile     [in] A pointer to the start of file in memory (cached file)
 * @param cachedFileSize [in] The total filesize in bytes
 * @return '1' when successful; '0' when an error occured
 */
static int decryptFileInAdvance(t_U8* const cachedFile, t_S32 const cachedFileSize)
{
    int result;

    if (cachedFileSize > AES128_KEYSIZE) {
        //Skip the header of the file
        result = appaes_aes128DecryptCbc(&cachedFile[AES128_KEYSIZE],
                                         &cachedFile[AES128_KEYSIZE],
                                         cachedFileSize - AES128_KEYSIZE);
    } else {
        result = OMN_EXIT_FAILURE;
    }

    return result;
}

/**
 * Provides the ECB decrypted file contents to the caller
 * @param buf [out] Buffer to store the results in
 * @param len [in]  Length of the requested data
 * @return Number of byes put in the buffer
 */
static int ecbDecrypt(char* buf, int len)
{
    int success;
    int const reqLen = len;
    int const originalFilePos = filepos;
    int const originalFileMax = filemax;
    int const fileBlockPos = 1 + (filepos / fileDecrypt.keyLength);
    int const blockOffset = filepos % fileDecrypt.keyLength;

    encryption_mode = 0;
    filemax = encfilemax;
    filepos += fileDecrypt.keyLength; //First block is file encryption header so skip this

    //TODO We can optimize by combining the three steps. We can also cache the last decrypted block so it is not
    //required to decrypt a block for every read. If cached, we can skip the block read and decrypt
    if (blockOffset != 0) {
        int const BytesToCopy = (int) ((blockOffset + len) <= fileDecrypt.keyLength) ? len : fileDecrypt.keyLength - blockOffset;

        filepos = fileBlockPos * fileDecrypt.keyLength;
        success = grub_read(fileDecrypt.inputBlock, fileDecrypt.keyLength);
        GRUB_ASSERT(success != 0);
        success = fileDecrypt.decryptEngine(fileDecrypt.inputBlock, fileDecrypt.decryptedBlock, fileDecrypt.keyLength);
        if (success != 0) {
            goto dec_error;
        }
        memcpy(buf, &fileDecrypt.decryptedBlock[blockOffset], BytesToCopy);
        buf += BytesToCopy;
        len -= BytesToCopy;
    }

    while (len >= fileDecrypt.keyLength) {
        success = grub_read(fileDecrypt.inputBlock, fileDecrypt.keyLength);
        GRUB_ASSERT(success != 0);
        success = fileDecrypt.decryptEngine(fileDecrypt.inputBlock, fileDecrypt.decryptedBlock, fileDecrypt.keyLength);
        if (success != 0) {
            goto dec_error;
        }
        memcpy(buf, fileDecrypt.decryptedBlock, fileDecrypt.keyLength);
        buf += fileDecrypt.keyLength;
        len -= fileDecrypt.keyLength;
    }

    if (len > 0) {
        success = grub_read(fileDecrypt.inputBlock, fileDecrypt.keyLength);
        GRUB_ASSERT(success != 0);
        success = fileDecrypt.decryptEngine(fileDecrypt.inputBlock, fileDecrypt.decryptedBlock, fileDecrypt.keyLength);
        if (success != 0) {
            goto dec_error;
        }
        memcpy(buf, fileDecrypt.decryptedBlock, len);
        buf += len;
        len -= len;
    }
    GRUB_ASSERT(len == 0); //Sanity check
    filepos = originalFilePos + reqLen;
    filemax = originalFileMax;
    encryption_mode = AES_ECB_MODE;
    return reqLen;

dec_error :
    UPDATE_LOG_ERR(ERR_BAD_FILE_DEC_PREPARE);
    return 0;
}

/**
 * Provides the CBC decrypted file contents to the caller
 * @param buf [out] Buffer to store the results in
 * @param len [in]  Length of the requested data
 * @return Number of byes put in the buffer
 */
static int cbcDecrypt(char* buf, int len)
{
    int bytesRead;

    filepos += fileDecrypt.keyLength; //Correct for the file header (which should be 'fileDecrypt.keyLength' in size
    filemax += fileDecrypt.keyLength;
    encryption_mode = 0;
    bytesRead = grub_read(buf, len); //As we use CBC, we assume the uncrypted file is avaiable via 'grub_read' to supply us the data
    encryption_mode = AES_CBC_MODE;
    filepos -= fileDecrypt.keyLength;
    filemax -= fileDecrypt.keyLength;

    return bytesRead;
}

/* ================ Global functions (Sub-system scope) ============================================================= */
/**
 * This function should be called to read a part of the encrypted file
 * @pre Use function @p filedecr_checkFileHeaderForEnc() to check if the file is encrypted. Only use @p filedecr_read()
 * if the file is encrypted
 * @note @p filemax, filepos are used for indexing the encrypted file
 * @param buf   [in] The buffer to store the requested decrypted data
 * @param len   [in] The length of data requested
 * @return
 */
int filedecr_read(char* buf, int len)
{
    int bytesRead = 0;

    if (fileDecrypt.keyLength > 0) {
        switch (encryption_mode) {
            case AES_ECB_MODE :
                bytesRead = ecbDecrypt(buf, len);
                break;

            case AES_CBC_MODE :
                bytesRead = cbcDecrypt(buf, len);
                break;

            default:
                UPDATE_LOG_ERR(ERR_UNSUPPORTED_ENC_METHOD);
                bytesRead = 0;
                break;
        }
    }
    return bytesRead;
}

/**
 * Check if a file is encrypted by decoding the file header and look for specific data
 * @param cachedFile [in] A pointer to a buffer containing the whole! file. This is required for AES CBC
 * @param cachedFileSize [in] The total size of the cached file
 * @return 1 if the file is encrypted; 0 if it is not
 */
int filedecr_evaluateAndProcess(t_U8* const cachedFile, t_S32 const cachedFileSize)
{
    t_aesFileHdr fileHeader;
    int rb;

    encryption_mode = 0;
    encfilemax = 0;
    filepos = 0;

    if (filemax >= (int) sizeof(fileHeader)) {
        rb = grub_read((char *) &fileHeader, sizeof(fileHeader));
        filepos = 0;
        if ((rb == 0) || (rb != (int) sizeof(fileHeader))) {
            UPDATE_LOG_ERR(ERR_READ);
            return 0;
        }

        if ((fileHeader.magic      == AES_FILE_HDR_MAGIC) &&
            (fileHeader.version    >= AES_ENC_FILE_HDR_VERSION) &&
            (fileHeader.aesKeySize == AES128_KEYSIZE)) {
#ifdef DEBUG
            printf("Encrypted file header with:\n");
            printf("magic: 0x%x\n", fileHeader.magic);
            printf("version: 0x%x\n", fileHeader.version);
            printf("AES mode: 0x%x\n", fileHeader.aesMode);
            printf("keySize: %d\n", fileHeader.aesKeySize);
            printf("fileSize: %d\n", fileHeader.fileSize);
#endif

            if ((filemax >= AES128_KEYSIZE) && ((filemax % AES128_KEYSIZE) == 0)) {
#ifdef SUPPORT_AES_ECB
                if (fileHeader.aesMode == AES_ECB_MODE) {
                    setDecryptionEngine(appeas_aes128DecryptEcb, AES128_KEYSIZE);
                    encfilemax      = filemax;
                    filemax         = fileHeader.fileSize;
                    encryption_mode = AES_ECB_MODE;
                } else
#endif
                if ((fileHeader.aesMode == AES_CBC_MODE) && (cachedFile != NULL) && (cachedFileSize > 0)) {
                    int status;

                    encfilemax                = filemax;
                    filemax                   = fileHeader.fileSize;
                    fileDecrypt.keyLength     = AES128_KEYSIZE;
                    fileDecrypt.decryptEngine = NULL;
                    status                    = decryptFileInAdvance(cachedFile, cachedFileSize);
                    if (status != OMN_EXIT_SUCCESS) {
                        UPDATE_LOG_ERR(ERR_DECRYPT_IN_ADVANCE);
                        return 0;
                    }
                    encryption_mode = AES_CBC_MODE;
                } else {
                    if (fileHeader.aesMode == AES_ECB_MODE) {
#ifdef DEBUG
                        printf("AES128-ECB not supported anymore\n");
#endif
                        UPDATE_LOG_ERR(ERR_UNSUPPORTED_ENC_METHOD);
                        return 0;
                    } else {
#ifdef DEBUG
                        printf("File is encrypted with AES128-CBC but we do not have the file in memory\n");
#endif
                        UPDATE_LOG_ERR(ERR_CACHE_ISSUE);
                        return 0;
                    }
                }
            } else {
                UPDATE_LOG_ERR(ERR_WRONG_ENC_FILE_SIZE);
                return 0;
            }
        } else if (fileHeader.magic == AES_FILE_HDR_MAGIC) {
#ifdef DEBUG
            printf("Encrypted file header with:\n");
            printf("magic: 0x%x\n", fileHeader.magic);
            printf("version: 0x%x\n", fileHeader.version);
#endif

            UPDATE_LOG_ERR(ERR_UNSUPPORTED_ENC_METHOD);
            return 0;
        } else {
#ifdef DEBUG
            printf("Encrypted file header with:");
            printf("magic: 0x%x\n", fileHeader.magic);
#endif
        }
    }
    UPDATE_LOG_ERR(0);
    return 1;
}

/**
 * Removes the AES decryption key from memory and clears the encrypted file indication
 */
void filedecr_terminate(void)
{
    setDecryptionEngine(NULL, 0);
    encryption_mode = 0;
}

/* ================ Global functions (Global system scope) ======================================================= *//**
 @endinternal
 *//* ================================================================================================================ */

/* ================ End of file ===================================================================================== */
