/* =====================================================================================================================
 (c) Copyright OMRON Europe BV 2016
 All Right Reserved
 OMRON Development & Engineering
 ================================================================================================================== *//**
 @file

  This file is part of Omron Secure Boot (OSB).

  Omron Secure Boot (OSB) is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Omron Secure Boot (OSB) is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Omron Secure Boot (OSB).  If not, see <http://www.gnu.org/licenses/>.


 Implements logic to load and verify a '.sig' file for the provided filename, calculate the SHA of the file and compare
 it with the RSA verified .sig contents. It will let the caller know if the requested file is validated or not.

 @internal
 *//* ================================================================================================================ */

/* ================ Global include files ============================================================================ */

/* ================ Sub-system include files ======================================================================== */
#include "shared.h"
#include "file_verify.h"
#include "rsa_cfg.h"
#include "types.h"
#include "sha2.h"
#include "app_rsa.h"
#include "utils.h"

/* ================ Local constant definitions ====================================================================== */

#define SIGNED_SHA2_FILE_EXT    ".sig"
#define MAX_FILENAME_LENGTH     (1500)

/* ================ Local macro definitions ========================================================================= */

/* ================ Local type definitions ========================================================================== */

/* ================ Local data definitions ========================================================================== */

/* ================ Local functions ================================================================================= */
/**
 * OSB commands can have parameters after the filename; this command will distill the filename so the correct .sig-file
 * can be loaded
 * @param filename  [in] The full filename of the file to check (without the .sig extension)
 * @return The the index after the root filename (just before the parameters)
 */
static int getRootFilename(char const * filename) {
    int i = 0;

    while ((*filename != 0) && (*filename != ' ')) {
        filename++;
        i++;
    }

    return i;
}

/**
 * Will load the contents of the .sig-file based on the provided @p filename and decrypts it with the public RSA
 * key
 * The 20 byte hash will be provided to the caller for further checking
 * @param filename   [in] The file to check (this function will automatically open the file <filename>.sig
 * @param sha2Value  [out] A pointer with at least @p SHA2_DWORD_SIZE in size to receive the SHA2 hash from the file
 * @return 1 if successful
 */
static int loadShaForFile(char* const filename, t_U32* const shaValue) {
    char const * const signedShaFnExt = SIGNED_SHA2_FILE_EXT;
    char shaFilename[MAX_FILENAME_LENGTH];
    t_U8 signedHash[RSA_KEYLEN];
    t_U8 shaHash[RSA_KEYLEN];
    int success = 0;
    char* currStrPos = shaFilename;

    if (filename != NULL) {
        int const filenameLength = getRootFilename(filename);

        if (filenameLength < ((int)sizeof(shaFilename) + grub_strlen(signedShaFnExt))) {
            int result;

            (void) memcpy(currStrPos, filename, getRootFilename(filename));
            currStrPos += getRootFilename(filename);
            (void) memcpy(currStrPos, signedShaFnExt, grub_strlen(signedShaFnExt));
            currStrPos += grub_strlen(signedShaFnExt);
            *currStrPos = '\0';

            result = grub_open(shaFilename);
            if (result != 0) {
                int const bytesToRead = sizeof(signedHash);

                result = grub_read((char*) signedHash, bytesToRead);
                if (result == bytesToRead) {
#ifdef DEBUG
                    grub_printf("File contents of .sig file: ");
                    utils_printAsHexStr((t_U8*)signedHash, sizeof(signedHash));
                    grub_printf("\n");
#endif
                    int shaHashSize = sizeof(shaHash);
                    int const status = apprsa_rsaVerify(signedHash, sizeof(signedHash), shaHash, &shaHashSize);

                    if ((status == OMN_EXIT_SUCCESS) && (shaHashSize >= (SHA2_DWORD_SIZE * (int)sizeof(t_U32)))) {
#ifdef DEBUG
                    grub_printf("RSA verified hash (original file hash): ");
                    utils_printAsHexStr((t_U8*)shaHash, sizeof(shaHash));
                    grub_printf("\n");
#endif
                        (void) memcpy(shaValue, shaHash, SHA2_DWORD_SIZE * sizeof(t_U32));
                        success = 1;
                    } else {
#ifdef DEBUG
                        grub_printf("RSA verify error\n");
#endif
                        success = 0;
                    }
                    apprsa_destroyKey();
#ifdef DEBUG
                    grub_printf("\n");
                } else {
                    grub_printf("Sign file does not contain a valid signed SHA2 hash\n\n");
#endif
                }
                grub_close();
            } else {
#ifdef DEBUG
                grub_printf("Could not open signed hash %s\n", shaFilename);
#endif
            }
        }
    }

    return success;
}

/**
 * Calculates the SHA hash for the given file(-name) and puts the result in @p shaValue
 * @param filename   [in] The file to check, can be @p NULL (if @p NULL this function assumes the file is already
 *                        opened somewhere else
 * @param shaValue  [out] A pointer with at least @p SHA2_DWORD_SIZE in size to receive the SHA hash from the file
 * @return 1 if successful
 */
static int calcShaForFile(char* const filename, t_U32* const shaValue) {
    SHA512_CTX shaCtx;
    int bytesRead;
    int status = 0;
    sha2_byte shaBuf[SHA_BUF_SIZE];
    int result;

    read_raw = 1;

    if (filename != NULL) {
        result = grub_open(filename);
    } else {
        //Someone else already opened the file for us so we don't have to do it anymore
        result = 1;
    }

    if (result != 0) {
        SHA512_Init(&shaCtx);
        (void) memset(shaBuf, 0x00, sizeof(shaBuf));

        filepos = 0;
        bytesRead = grub_read((char*) shaBuf, sizeof(shaBuf));
        while (bytesRead == (int)sizeof(shaBuf)) {
            SHA512_Update(&shaCtx, shaBuf, bytesRead);
            bytesRead = grub_read((char*) shaBuf, sizeof(shaBuf));
        }
        if (bytesRead > 0) {
            SHA512_Update(&shaCtx, shaBuf, bytesRead);
        }

        SHA512_Final((sha2_byte*)shaValue, &shaCtx);
        filepos = 0;
        if (filename != NULL) {
            grub_close();
        }
        status = 1;
    }
    read_raw = 0;

    return status;
}

/* ================ Global functions (Sub-system scope) ============================================================= */
/**
 * Returns a buffer containing the RSA decrypted file contens of <fn>.sig
 * @param fn [in] The original filename to get the SHA2 hash for
 * @return A pointer to the buffer containing the RSA decrypted SHA2 hash
 */
t_U32* fileverify_getVerifiedHashForFile(char* const fn)
{
    int          shaSuccess;
    static t_U32 unverifiedHash[SHA2_DWORD_SIZE];
    t_U32*       result = NULL;

    load_sha_for_file = 0;
    retry_open        = 0;
    shaSuccess        = loadShaForFile(fn, unverifiedHash);

    if (shaSuccess != 0) {
        result = unverifiedHash;
    } else {
        UPDATE_LOG_WARN_EXT(ERR_FILE_VERIFICATION, fn);
        result = NULL;
    }

    load_sha_for_file = 1;
    retry_open        = 1;

    return result;
}

/**
 * Main function to validate file @p fn using a signed .sig-file
 * @param fn              [in] The file to check (without the .sig extension)
 * @param optSha2FileHash [in] An optional SHA2 hash of the file retrieved in another way. If not available, set to @p NULL
 * @return 1 when successful. Please use @p errnum when the return value is 0 to get detailed info
 */
int fileverify_checkHashForFile(char *fn, t_U32 const * const optSha2FileHash)
{
    t_U32 unverifiedHash[SHA2_DWORD_SIZE];
    t_U32 fileHash[SHA2_DWORD_SIZE];
    int shaSuccess;

    load_sha_for_file       = 0;
    int pre_compressed_file = compressed_file;
    int pre_filemax         = filemax;
    compressed_file         = 0;
    retry_open              = 0;

    if (optSha2FileHash != NULL) {
        (void)memcpy(unverifiedHash, optSha2FileHash, sizeof(unverifiedHash));
        shaSuccess = 1;
    } else if (fn != NULL) {
        shaSuccess = loadShaForFile(fn, unverifiedHash);
    } else {
#ifdef DEBUG
        printf("No optional SHA2 hash is provided and no filename is provided\n");
#endif
        shaSuccess = 0;
    }

    if (shaSuccess != 0) {
#ifdef DEBUG
        if (fn != NULL) {
            printf("SHA2 hash file for file '%s' correctly loaded: \n", fn);
        } else {
            printf("Optional SHA2 hash correctly provided: \n");
        }
        utils_printAsHexStr((t_U8*)unverifiedHash, sizeof(unverifiedHash));
        printf("\n");
#endif
        shaSuccess = calcShaForFile(fn, fileHash);
        if (shaSuccess != 0) {
            int const cmpResult = memcmp((char*)fileHash, (char*)unverifiedHash, sizeof(fileHash));

#ifdef DEBUG
            if (fn != NULL) {
                printf("Calculated SHA2 hash for file '%s': \n", fn);
            } else {
                printf("Calculated SHA2 hash for externally opened file: \n", fn);
            }
            utils_printAsHexStr((t_U8*)fileHash, sizeof(fileHash));
            printf("\n");
#endif

            if (cmpResult == 0) {
#ifdef DEBUG
                printf("File verified successfully and allowed for use in OSB\n");
#endif
            } else {
                load_sha_for_file = 1;
                retry_open        = 1;
#ifdef DEBUG
                if (fn != NULL) {
                    printf("SHA2 hash verification for file '%s' failed\n", fn);
                } else {
                    printf("SHA2 hash verification for externally opened file failed\n");
                }
#endif
                UPDATE_LOG_WARN_EXT(ERR_FILE_VERIFICATION, fn);
                goto finish;
            }
        } else {
            load_sha_for_file = 1;
            retry_open        = 1;
#ifdef DEBUG
            if (fn != NULL) {
                printf("Could not correctly open file '%s' for SHA2 verification\n", fn);
            } else {
                printf("Could not correctly open SHA2 verification file\n");
            }
#endif
            UPDATE_LOG_WARN_EXT(ERR_FILE_VERIFICATION, fn);
            goto finish;
        }
    } else {
        load_sha_for_file = 1;
        retry_open        = 1;
#ifndef DEBUG
        if (fn != NULL) {
            printf("SHA2 verification of file '%s' failed\n", fn);
        } else {
            printf("SHA2 verification failed\n", fn);
        }
        UPDATE_LOG_WARN_EXT(ERR_FILE_VERIFICATION, fn);
        goto finish;
#endif
    }
    load_sha_for_file = 1;
    retry_open        = 1;

    // restore compression settings
    compressed_file = pre_compressed_file;
    filemax = pre_filemax;
    return 1;

finish:
    // restore compression settings
    compressed_file = pre_compressed_file;
    filemax = pre_filemax;
    return 0;
}


/* ================ Global functions (Global system scope) ======================================================= *//**
 @endinternal
 *//* ================================================================================================================ */

/* ================ End of file ===================================================================================== */
