/* trace.c - tracing and error reporting related functionallity */
/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 1999,2000,2001,2002,2004  Free Software Foundation, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* ================ Global include files ============================================================================ */

#include <shared.h>
#include "term.h"

/* ================ Sub-system include files ======================================================================== */

/* ================ Local constant definitions ====================================================================== */

/* ================ Local macro definitions ========================================================================= */

#define LOG_BUFFER_SIZE (0x4000)

/* ================ Local type definitions ========================================================================== */

/* ================ Local data definitions ========================================================================== */

/*
 * Integer indicating the highest added message type
 */
static int highestTypeLogged = LOG_INFO;

/*
 * buffer to store log data
 */
static char logBuffer[LOG_BUFFER_SIZE+1] = {0};

/*
 * Type log log messages
 */
static char const * const logType_list[] =
{
 [LOG_INFO] = "Info",
 [LOG_WARN] = "Warning",
 [LOG_ERR] = "Error",
};

/* ================ Local functions ================================================================================= */

/**
 * Determine if log is full
 * @param void
 * @return boolean indicating if log is full
 */
static t_bool logFull(void) {
    return ((strlen(logBuffer) >= LOG_BUFFER_SIZE) ? TRUE : FALSE);
}

/**
 * Clear Log
 * @param void
 * @return void
 */
static void clearLog(void) {
    logBuffer[0] = '\0';
    highestTypeLogged = LOG_INFO;
}

/* ================ Global functions (Sub-system scope) ============================================================= */

/* ================ Global functions (Global system scope) ======================================================= */

/**
 * Update Log
 * @param messageType [in] Type of message
 * @param assignedErrNum [in] Error number of this error message
 * @param argStr [in] Additional description string of the message
 * @param callerFile [in] code filename of the caller
 * @param callerLine [in] code line number of the caller
 * @param callerFunc [in] function name of the caller
 * @return void
 */
void updateLog(LogType const messageType, int const assignedErrNum, char const* const argStr, char const* const callerFile, int const callerLine, char const* const callerFunc) {
    int logBufferLen = strlen(logBuffer);
    const t_bool isLogFull = logFull();
    errnum=assignedErrNum;

    // is there space left
    if ((errnum > ERR_NONE) && (errnum < MAX_ERR_NUM)) {
        // store highest priority message type
        if (messageType > highestTypeLogged) {
            highestTypeLogged = messageType;
        }
        if (ISFALSE(isLogFull)) {
            if (argStr != 0 && *argStr != '\0') {
                snprintf(logBuffer+logBufferLen, LOG_BUFFER_SIZE-logBufferLen, " - %s: %s (0x%x)\n   Details: '%s'\n   %s:%d %s\n\n", logType_list[messageType], err_list[errnum], errnum, argStr, callerFile, callerLine, callerFunc);
            } else {
                snprintf(logBuffer+logBufferLen, LOG_BUFFER_SIZE-logBufferLen, " - %s: %s (0x%x)\n   %s:%d %s\n\n", logType_list[messageType], err_list[errnum], errnum, callerFile, callerLine, callerFunc);
            }
        }
        // report immediately if we have an error
        if (highestTypeLogged >= LOG_ERR) {
            reportLog(LOG_ERR);
#ifdef DEBUG
            getkey();
#else
            for(;;);
#endif
        }
    }
}

/**
 * Report Log
 * @param messageType [int] minimal report level
 * @return void
 */
void reportLog(LogType const messageType) {
    const t_bool isLogFull = logFull();
    // if the highest logged error is greater or equal to the priority we are looking for
    if (highestTypeLogged >= messageType) {
        printf("Please consult the manual to determine the next steps.\n");
        printf("If you need to report the problem, please include the following information in the report:\n\n%s", logBuffer);
        if (ISTRUE(isLogFull)) {
            printf("No space left to print additional messages\n");
        }
        printf("Release: %s", PACKAGE_STRING);
        clearLog();
    }
}
/* ================================================================================================================ */

/* ================ End of file ===================================================================================== */
