/* =====================================================================================================================
 (c) Copyright OMRON Europe BV 2016
 All Right Reserved
 OMRON Development & Engineering
 ================================================================================================================== *//**
 @file

  This file is part of Omron Secure Boot (OSB).

  Omron Secure Boot (OSB) is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Omron Secure Boot (OSB) is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Omron Secure Boot (OSB).  If not, see <http://www.gnu.org/licenses/>.

 Linux command line tool to generate an decrypted version of an input file together with a key-file

 @note Only AES128 is supported now

 @internal
 *//* ================================================================================================================ */

/* ================ Global include files ============================================================================ */
#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <time.h>
#include "aes.h"
#include "types.h"
#include "misc.h"

/* ================ Sub-system include files ======================================================================== */

/* ================ Local constant definitions ====================================================================== */
#define AES_FILE_HDR_MAGIC          (0x05ECB007UL)
#define AES_ENC_FILE_HDR_VERSION    (0x0110U)  //1.1
#define AES_CBC_MODE                (0x0002U)
#define REQ_ARGS_CNT                (5)         //filename + the three arguments
#define REQ_INFO_ARGS_CNT           (4)         //Filename + -i + infilename argument

#define MAX_FILE_SIZE               (300 * 1024 * 1024)
#define AES_KEY_BUF_LEN             (64)

#define KEYFILE_NOT_FOUND           (1)
#define INFILE_NOT_FOUND            (2)
#define INFILE_SIZE_MISMATCH        (3)
#define INFILE_HDR_VERSION_ERROR    (4)
#define INFILE_HDR_AES_MODE_ERROR   (5)
#define KEYFILE_KEYLENGTH_ERROR     (6)
#define INFILE_NO_HEADER            (7)
#define INFILE_TOO_SMALL            (8)
#define INFILE_CORRUPT_HEADER       (9)
#define SIGNFILE_OPEN_ERROR         (10)
#define OUTFILE_WRITE_ERROR         (11)
#define UNSUPPORTED_KEYLENGTH       (12)
#define INTERNAL_ERROR              (13)
#define OUTFILE_CREATE_ERROR        (14)
#define FILE_TOO_LARGE              (15)

/* ================ Local macro definitions ========================================================================= */
#define LOBYTE(w)  ((t_U8) ((unsigned)(w)       & 0xFFU))
#define HIBYTE(w)  ((t_U8)(((unsigned)(w) >> 8) & 0xFFU))

/* ================ Local type definitions ========================================================================== */
typedef struct {
    t_U32 magic;
    t_U16 version;
    t_U16 aesMode;
    t_U16 aesKeySize;
    t_U32 fileSize;
}PACKED_BYTE_ALIGN(T_AES_FILE_HDR);

/* ================ Local data definitions ========================================================================== */
static char* errMsg[] ={
     [KEYFILE_NOT_FOUND]        = "Could not open the key-file",
     [INFILE_NOT_FOUND ]        = "Could not open the input file",
     [INFILE_SIZE_MISMATCH]     = "The input file does not have the expected file size for correct decryption",
     [INFILE_HDR_VERSION_ERROR] = "The input file encryption header version is not supported by this version of the application",
     [INFILE_HDR_AES_MODE_ERROR]= "The input file uses an unsupported AES mode",
     [KEYFILE_KEYLENGTH_ERROR]  = "The key-file length does not match the key length in the file encryption header",
     [INFILE_NO_HEADER]         = "The input file does not seem to be an encrypted Omron Secure Boot file",
     [INFILE_TOO_SMALL]         = "The input file is not big enough to be an encrypted Omron Secure Boot file",
     [INFILE_CORRUPT_HEADER]    = "The input file seems to have a corrupted header",
     [OUTFILE_CREATE_ERROR]     = "Could not create the output file",
     [OUTFILE_WRITE_ERROR]      = "Could not write to the output file",
     [UNSUPPORTED_KEYLENGTH]    = "The key-file contains an unsupported key length",
     [INTERNAL_ERROR]           = "Oeps, something went really wrong",
     [FILE_TOO_LARGE]           = "The file to encrypt is too large to hold in the buffer"};

static t_U8 iv[AES128_KEYLEN] = {0x61, 0xff, 0x1a, 0x50, 0x01, 0xae, 0xe5, 0x9b, 0x95, 0x66, 0x4d, 0xde, 0x85, 0xe0, 0x99, 0xaa};
static t_U8 decBuffer[MAX_FILE_SIZE];

static int info  = 0;

static char*            optstring = "hiv";
static struct option    longopts[] = {
    { "help", no_argument, 0, 'h' },
    { "info", no_argument, 0, 'i' },
    { "version", no_argument, 0, 'v' },
    { 0 } };

/* ================ Local functions ================================================================================= */
/**
 * Prints a help message to the standard console
 * @param doStandardHelp
 */
static void help(int const doStandardHelp)
{
    if (doStandardHelp)
        fprintf(stderr, "Try ``decrypt_file --help'' for more information.\n");
    else
        printf("decrypt_file (OSB " VERSION ")\n"
                        "Usage: decrypt_file [KEYFILE] [KEY_IDX] [INFILE] [OUTFILE]\n"
                        "Usage: decrypt_file [-i|--info] [INFILE]\n"
                        "Decrypts the file [INFILE] using the provided [KEYFILE] with key at index [KEY_IDX] and writes the result to [OUTFILE].\n"
                        "\n"
                        "-i, --info                 prints the file-header of the provided file; does not decrypt the file\n"
                        "-h, --help                 display this help and exit\n"
                        "-v, --version              output version information and exit.\n"
                        "\n"
                        "Report bugs to <info@omron.com>.\n");

    exit(doStandardHelp);
}

/**
 * Prints the Omron encryption header to the console
 * @param filename  [in] The filename for which this header is provided
 * @param aesFileHdr[in] The file header information of file @p filename
 */
static void printInfo(char const * const filename, T_AES_FILE_HDR const * const aesFileHdr)
{
    printf("Omron Secure Boot header information for file '%s':\n"
                    "    Version              : %x.%x\n"
                    "    AES mode             : %u\n"
                    "    Original file size   : %u\n"
                    "    Encryption key length: %u\n"
                    "\n", filename, HIBYTE(aesFileHdr->version), LOBYTE(aesFileHdr->version),
                    (unsigned int)aesFileHdr->aesMode, (unsigned int) aesFileHdr->fileSize,
                    (unsigned int) aesFileHdr->aesKeySize);
}

/**
 * Adds the proper encryption file header to the output file for OSB to parse
 * @param outFile       [in] File stream of the output file
 * @param inputFileSize [in] The input file file-size in bytes
 * @return 0 when successful
 */
static int decodeFileEncHeader(FILE* const inFile, long int const inputFileSize, T_AES_FILE_HDR* const aesFileHdr)
{
    int status;
    int bytesRead;

    if (inputFileSize >= (int)sizeof(*aesFileHdr)) {
        bytesRead = fread(aesFileHdr, 1, sizeof(*aesFileHdr), inFile);
        if (bytesRead == (int)sizeof(*aesFileHdr)) {
            if ((aesFileHdr->magic   == AES_FILE_HDR_MAGIC) &&
                (aesFileHdr->version == AES_ENC_FILE_HDR_VERSION) &&
                (aesFileHdr->aesMode == AES_CBC_MODE)) {
                status = OMN_EXIT_SUCCESS;
            } else if (aesFileHdr->version != AES_ENC_FILE_HDR_VERSION) {
                status = INFILE_HDR_VERSION_ERROR;
            } else if (aesFileHdr->aesMode != AES_CBC_MODE) {
                status = INFILE_HDR_AES_MODE_ERROR;
            } else {
                status = INFILE_NO_HEADER;
            }
        } else {
            status = INTERNAL_ERROR;
        }
    } else {
        status = INFILE_TOO_SMALL;
    }

    return status;
}

/**
 * Encrypts the input file with the encryption key and writes it to the output file
 * @param inFile    [in]  File stream of the input file
 * @param encKey    [in]  Byte array containing the encryption key
 * @param keySize   [in]  The size, in bytes, of the encryption key
 * @param outFile   [out] File stream to which the encrypted data is written to
 * @return 0 when successful
 */
static int decryptData(FILE* const                  inFile,
                       long int const               inputFileSize,
                       t_U8* const                  decKey,
                       int const                    keySize,
                       T_AES_FILE_HDR const * const fileHdr,
                       FILE* const                  outFile)
{
    int status = OMN_EXIT_SUCCESS;

    if ((inputFileSize % keySize) == 0) {
        (void) fseek(inFile, keySize, SEEK_SET); //Back to the beginning of the input file, skipping the header

        if (keySize == (int)fileHdr->aesKeySize) {
            if (keySize == AES128_KEYLEN) {
                int  dummyReadCount;
                t_U8 dummy;
                int  bytesRead;

                bytesRead      = fread(decBuffer, 1, sizeof(decBuffer)-AES128_KEYLEN, inFile);
                dummyReadCount = fread(&dummy, 1, sizeof(dummy), inFile);

                if (dummyReadCount == 0) {
                    if ((bytesRead >= fileHdr->fileSize) && ((bytesRead % AES128_KEYLEN) == 0)) {
                        int bytesWritten;

                        aes_cbc_decrypt(decBuffer, decBuffer, bytesRead, decKey, iv);

                        bytesWritten = (int)fwrite(decBuffer, 1, (size_t)fileHdr->fileSize, outFile);
                        if (bytesWritten != fileHdr->fileSize) {
                            status = OUTFILE_WRITE_ERROR;
                        }
                    } else {
                        status = INFILE_SIZE_MISMATCH;
                    }
                } else {
                    status = FILE_TOO_LARGE;
                }
            } else {
                status = UNSUPPORTED_KEYLENGTH;
            }
        } else {
            status = KEYFILE_KEYLENGTH_ERROR;
        }
    } else {
        status = INFILE_SIZE_MISMATCH;
    }

    return status;
}

/**
 * Preparation function to encrypt an input file with the provided key-file and generate the encrypted output file
 * @param keyfile   [in] Filename of the key-file (including path)
 * @param keyIndex  [in] The AES key at index N, starting from 0
 * @param infile    [in] Filename of the input-file (including path)
 * @param outfile   [in] Filename of the output-file (including path)
 * @return 0 when successful
 */
static int decryptFile(char const * const keyfile, int const keyIndex, char const * const infile, char const * const outfile)
{
    t_U8            decKey[AES_KEY_BUF_LEN];
    int             keySize = sizeof(decKey);
    int             status;
    FILE*           inFile;
    long int        inFileSize;
    T_AES_FILE_HDR  aesFileHdr;

    inFile = fopen(infile, "r");
    if (inFile == NULL) {
        status = INFILE_NOT_FOUND;
        goto error;
    }

    (void) fseek(inFile, 0L, SEEK_END);
    inFileSize = ftell(inFile);
    (void) fseek(inFile, 0L, SEEK_SET);

    status = decodeFileEncHeader(inFile, inFileSize, &aesFileHdr);
    if (status != 0) {
        fclose(inFile);
        goto error;
    }

    if (info == 0) {
        FILE* outFile = fopen(outfile, "w+");

        if (outFile == NULL) {
            fclose(inFile);
            status = OUTFILE_CREATE_ERROR;
            goto error;
        }

        status = misc_getAesKey(keyfile, keyIndex, decKey, &keySize);

#ifdef DEBUG
        if (status == OMN_EXIT_SUCCESS) {
            printf("AES key ");
            misc_printKey((t_U8*) decKey, AES128_KEYLEN);
            printf("\n");
            printf("Index 0: %x, %d: %x", (t_U8) decKey[0], AES128_KEYLEN - 1,
                            (t_U8) decKey[AES128_KEYLEN - 1]);
            printf("\n");
        }
#endif
        if (status != 0) {
            fclose(inFile);
            fclose(outFile);
            goto error;
        }

        status = decryptData(inFile, inFileSize, decKey, keySize, &aesFileHdr, outFile);
        fclose(outFile);
    } else {
        printInfo(infile, &aesFileHdr);
    }

    fclose(inFile);

    error : return status;
}

/* ================ Global functions (Sub-system scope) ============================================================= */

/* ================ Global functions (Global system scope) ======================================================= *//**
 @endinternal
 *//* ================================================================================================================ */

/**
 * Main entry function of the application
 * @param argc [in] number of arguments
 * @param argv [in] values of arguments
 * @return successful completion
 */
int main(int argc, char *argv[])
{
    int c;

    srand((unsigned int)time(NULL));

    do {
        c = getopt_long(argc, argv, optstring, longopts, 0);
        switch ((char)c) {
            case (char)EOF:
                break;

            case 'h':
                help(0);
                break;

            case 'v':
                printf("decrypt_file (OSB " VERSION ")\n");
                exit(0);
                break;

            case 'i':
                info = 1;
                break;

            default:
                help(1);
                break;
        }
    } while (c != EOF);

    if (optind < argc) {
        if ((info != 0) && (argc == REQ_INFO_ARGS_CNT)) {
            char const * const infile = argv[optind++];
            int const keyIndex        = (int) atoi(argv[optind++]);
            int const status          = decryptFile(NULL, keyIndex, infile, NULL);

            if (status != 0) {
                fprintf(stderr, "ERROR: '%s'\n", errMsg[status]);
                exit(1);
            }
        } else if ((info == 0) && (argc == REQ_ARGS_CNT)) {
            char const * const keyfile = argv[optind++];
            int const keyIndex         = (int) atoi(argv[optind++]);
            char const * const infile  = argv[optind++];
            char const * const outfile = argv[optind++];
            int const status = decryptFile(keyfile, keyIndex, infile, outfile);

            if (status != 0) {
                fprintf(stderr, "ERROR: '%s'\n", errMsg[status]);
                exit(1);
            }
        } else {
            fprintf(stderr,
                            "Please provide the required parameters: 'decrypt_file [KEYFILE] [KEY_IDX] [INFILE] [OUTFILE]' or decrypt_file -i [INFILE]' \n");
            exit(1);
        }
    } else {
    }

    return 0;
}

/* ================ End of file ===================================================================================== */
