/* =====================================================================================================================
 (c) Copyright OMRON Europe BV 2016
 All Right Reserved
 OMRON Development & Engineering
 ================================================================================================================== *//**
 @file

  This file is part of Omron Secure Boot (OSB).

  Omron Secure Boot (OSB) is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Omron Secure Boot (OSB) is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Omron Secure Boot (OSB).  If not, see <http://www.gnu.org/licenses/>.

 Linux command line tool to generate an encrypted version of an input file together with a key-file

 @note Only AES128 is supported now

 @internal
 *//* ================================================================================================================ */

/* ================ Global include files ============================================================================ */
#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <time.h>
#include "types.h"
#include "aes.h"
#include "misc.h"

/* ================ Sub-system include files ======================================================================== */

/* ================ Local constant definitions ====================================================================== */
#define AES_FILE_HDR_MAGIC          (0x05ECB007UL)
#define AES_ENC_FILE_HDR_VERSION    (0x0110U)  //1.1
#define AES_CBC_MODE                (0x0002U)
#define REQ_ARGS_CNT                (5)         //filename + the three arguments

#define MAX_FILE_SIZE               (300 * 1024 * 1024)
#define AES_KEY_BUF_LEN             (64)

#define KEYFILE_NOT_FOUND           (1)
#define INFILE_NOT_FOUND            (2)
#define OUTFILE_CREATE_ERROR        (3)
#define OUTFILE_WRITE_ERROR         (4)
#define UNSUPPORTED_KEYLENGTH       (5)
#define FILE_TOO_LARGE              (6)

/* ================ Local macro definitions ========================================================================= */
#define CEIL(x, y) (((x + (y - 1)) / y) * y)

/* ================ Local type definitions ========================================================================== */

/* ================ Local data definitions ========================================================================== */
static char* errMsg[] = {
      [KEYFILE_NOT_FOUND]       = "Could not open the key-file",
      [INFILE_NOT_FOUND]        = "Could not open the input file",
      [OUTFILE_CREATE_ERROR]    = "Could not create the output file",
      [OUTFILE_WRITE_ERROR]     = "Could not write to the output file",
      [UNSUPPORTED_KEYLENGTH ]  = "The key-file contains an unsupported key length",
      [FILE_TOO_LARGE]          = "The file to encrypt is too large to hold in the buffer"
};

static t_U8 iv[AES128_KEYLEN] = {0x61, 0xff, 0x1a, 0x50, 0x01, 0xae, 0xe5, 0x9b, 0x95, 0x66, 0x4d, 0xde, 0x85, 0xe0, 0x99, 0xaa};
static t_U8 encBuffer[MAX_FILE_SIZE];

static char*         optstring  = "hv";
static struct option longopts[] = {
    { "help", no_argument, 0, 'h' },
    { "version", no_argument, 0, 'v' },
    { 0 } };

/* ================ Local functions ================================================================================= */
/**
 * Prints a help message to the standard console
 * @param doStandardHelp
 */
static void help(int const doStandardHelp)
{
    if (doStandardHelp)
        fprintf(stderr, "Try ``encrypt_file --help'' for more information.\n");
    else
        printf("encrypt_file (OSB " VERSION ")\n"
                        "Usage: encrypt_file [KEYFILE] [KEY_IDX] [INFILE] [OUTFILE]\n"
                        "Encrypts the file [INFILE] using the provided [KEYFILE] and writes the result to [OUTFILE].\n"
                        "\n"
                        "-h, --help                 display this help and exit\n"
                        "-v, --version              output version information and exit.\n"
                        "\n"
                        "Report bugs to <info@omron.com>.\n");

    exit(doStandardHelp);
}

/**
 * Fills a buffer with random data
 * @param buf       [out] Buffer to receive random data
 * @param bufSize   [in]  The size of the buffer
 */
static void randomFill(t_U8* buf, int bufSize)
{
    while (bufSize > 0) {
        *buf = (t_U8)rand();
        buf++;
        bufSize--;
    }
}

/**
 * Adds the proper encryption file header to the output file for OSB to parse
 * @param outFile       [in] File stream of the output file
 * @param keyLength     [in] The AES encryption key length in bytes
 * @param inputFileSize [in] The input file file-size in bytes
 * @return 0 when successful
 */
static int addFileEncHeader(FILE* const outFile, int const keyLength, long int const inputFileSize)
{
    int status;
    t_U8 easFileHeader[keyLength];

    struct {
        t_U32 magic;
        t_U16 version;
        t_U16 aesMode;
        t_U16 aesKeySize;
        t_U32 fileSize;
    }PACKED_BYTE_ALIGN(*aesFileHdr)= (void*)easFileHeader;

    if (keyLength >= (int)sizeof(*aesFileHdr)) {
        randomFill(easFileHeader, sizeof(easFileHeader));
        aesFileHdr->magic       = AES_FILE_HDR_MAGIC;
        aesFileHdr->version     = AES_ENC_FILE_HDR_VERSION;
        aesFileHdr->aesMode     = AES_CBC_MODE;
        aesFileHdr->aesKeySize  = (t_U16)keyLength;
        aesFileHdr->fileSize    = (t_U32) inputFileSize;

        status = (int)fwrite(easFileHeader, 1, (size_t)sizeof(easFileHeader), outFile);
        if (status == (int)sizeof(easFileHeader)) {
            status = OMN_EXIT_SUCCESS;
        } else {
            status = OUTFILE_WRITE_ERROR;
        }
    } else {
        status = UNSUPPORTED_KEYLENGTH;
    }

    return status;
}

/**
 * Encrypts the input file with the encryption key and writes it to the output file
 * @param inFile    [in]  File stream of the input file
 * @param encKey    [in]  Byte array containing the encryption key
 * @param keySize   [in]  The size, in bytes, of the encryption key
 * @param outFile   [out] File stream to which the encrypted data is written to
 * @return 0 when successful
 */
static int encryptData(FILE* const inFile, t_U8* const encKey, int const keySize, FILE* const outFile)
{
    int haveWriteError = 0;
    int status = OMN_EXIT_SUCCESS;

    if (keySize == AES128_KEYLEN) {
        int  bytesRead;
        int  bytesWritten;
        t_U8 dummy;
        int  testReadCnt;

        bytesRead   = fread(encBuffer, 1, sizeof(encBuffer)-AES128_KEYLEN, inFile);
        testReadCnt = fread(&dummy, 1, sizeof(dummy), inFile);

        if (testReadCnt == 0) {
            aes_cbc_encrypt(encBuffer, encBuffer, CEIL(bytesRead, AES128_KEYLEN), encKey, iv);
            bytesWritten = (int)fwrite(encBuffer, 1, (size_t)CEIL(bytesRead, AES128_KEYLEN), outFile);
            if (bytesWritten == (int)CEIL(bytesRead, AES128_KEYLEN)) {
                status = OMN_EXIT_SUCCESS;
            } else {
                status = OUTFILE_WRITE_ERROR;
            }
        } else {
            status = FILE_TOO_LARGE;
        }
    } else {
        status = UNSUPPORTED_KEYLENGTH;
    }

    if (haveWriteError != 0) {
        status = OUTFILE_WRITE_ERROR;
    }

    return status;
}

/**
 * Preparation function to encrypt an input file with the provided key-file and generate the encrypted output file
 * @param keyfile   [in] Filename of the key-file (including path)
 * @param keyIndex  [in] The entry in the key-file, counting from 0
 * @param infile    [in] Filename of the input-file (including path)
 * @param outfile   [in] Filename of the output-file (including path)
 * @return 0 when successful
 */
static int encryptFile(char const * const keyfile, int const keyIndex, char const * const infile, char const * const outfile)
{
    t_U8     encKey[AES_KEY_BUF_LEN];
    int      keySize = sizeof(encKey);
    int      status;
    FILE*    outFile;
    FILE*    inFile;
    long int inFileSize;

    status = misc_getAesKey(keyfile, keyIndex, encKey, &keySize);
    if (status != OMN_EXIT_SUCCESS) {
        goto error;
    }

    outFile = fopen(outfile, "w+");
    if (outFile == NULL) {
        status = OUTFILE_CREATE_ERROR;
        goto error;
    }

    inFile = fopen(infile, "r");
    if (inFile == NULL) {
        fclose(outFile);
        status = INFILE_NOT_FOUND;
        goto error;
    }

    (void) fseek(inFile, 0L, SEEK_END);
    inFileSize = ftell(inFile);
    (void) fseek(inFile, 0L, SEEK_SET);

    status = addFileEncHeader(outFile, keySize, inFileSize);
    if (status != 0) {
        fclose(inFile);
        fclose(outFile);
        goto error;
    }

    status = encryptData(inFile, encKey, keySize, outFile);
    if (status != 0) {
        fclose(inFile);
        fclose(outFile);
        goto error;
    }

    fclose(inFile);
    fclose(outFile);

    error : return status;
}

/* ================ Global functions (Sub-system scope) ============================================================= */

/* ================ Global functions (Global system scope) ======================================================= *//**
 @endinternal
 *//* ================================================================================================================ */

/**
 * Main entry function of the application
 * @param argc [in] number of arguments
 * @param argv [in] values of arguments
 * @return successful completion
 */
int main(int argc, char *argv[])
{
    int c;

    srand((unsigned int)time(NULL));

    do {
        c = getopt_long(argc, argv, optstring, longopts, 0);
        switch ((char)c) {
            case (char)EOF:
                break;

            case 'h':
                help(0);
                break;

            case 'v':
                printf("encrypt_file (OSB " VERSION ")\n");
                exit(0);
                break;

            default:
                help(1);
                break;
        }
    } while (c != EOF);

    if (optind < argc) {
        if (argc == REQ_ARGS_CNT) {
            char const * const keyfile  = argv[optind++];
            int const          keyIndex = (int) atoi(argv[optind++]);
            char const * const infile = argv[optind++];
            char const * const outfile = argv[optind++];
            int const          status = encryptFile(keyfile, keyIndex, infile, outfile);

            if (status != 0) {
                fprintf(stderr, "ERROR: '%s'\n", errMsg[status]);
                exit(1);
            }
        } else {
            fprintf(stderr, "Please provide the required parameters: 'encrypt_file [KEYFILE] [KEY_IDX] [INFILE] [OUTFILE]'\n");
            exit(1);
        }
    } else {
    }

    return 0;
}

/* ================ End of file ===================================================================================== */
