/* =====================================================================================================================
 (c) Copyright OMRON Europe BV 2016
 All Right Reserved
 OMRON Development & Engineering
 ================================================================================================================== *//**
 @file

 This file is part of Omron Secure Boot (OSB).

 Omron Secure Boot (OSB) is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Omron Secure Boot (OSB) is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Omron Secure Boot (OSB).  If not, see <http://www.gnu.org/licenses/>.

 Linux command line tool to generate an encrypted version of an input file together with a key-file

 @note Only AES128 is supported now

 @internal
 *//* ================================================================================================================ */

/* ================ Global include files ============================================================================ */
#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <time.h>
#include "types.h"
#include "misc.h"
#include "aes.h"
#include "rsa_cfg.h"
#include "bigdigits.h"

/* ================ Sub-system include files ======================================================================== */

/* ================ Local constant definitions ====================================================================== */
#define REQ_ARGS_CNT                (7)

#define AES_OUTFILE_CREATE_ERROR    (1)
#define RSA_OUTFILE_CREATE_ERROR    (2)
#define AES_OUTFILE_WRITE_ERROR     (3)
#define RSA_OUTFILE_WRITE_ERROR     (4)
#define UNSUPPORTED_RSA_KEYLENGTH   (5)
#define UNSUPPORTED_AES_KEYLENGTH   (6)
#define PARAM_ERROR                 (7)

/* ================ Local macro definitions ========================================================================= */
#ifdef __ANALYSE__
#define PACKED_BYTE_ALIGN(type)        (type)
#else
#define PACKED_BYTE_ALIGN(type)        __attribute__ ((packed, aligned(1))) (type)
#endif

/* ================ Local type definitions ========================================================================== */

/* ================ Local data definitions ========================================================================== */
static char* errMsg[] = { [AES_OUTFILE_CREATE_ERROR] = "Could not create AES output file", [RSA_OUTFILE_CREATE_ERROR
                ] = "Could not create RSA output file",
                          [AES_OUTFILE_WRITE_ERROR] = "Could not write to AES output file", [RSA_OUTFILE_WRITE_ERROR
                                          ] = "Could not write to RSA output file",
                          [UNSUPPORTED_RSA_KEYLENGTH] = "The specified RSA key length is not a multiple of 4 bytes",
                          [UNSUPPORTED_AES_KEYLENGTH ] = "The specified AES key length is incorrect",
                          [PARAM_ERROR] = "Illegal parameter provided" };

static int quiet = 0;
static char* optstring = "hvq";
static struct option longopts[] = { { "help", no_argument, 0, 'h' },
                                    { "version", no_argument, 0, 'v' },
                                    { "quiet", no_argument, 0, 'q' },
                                    { 0 } };

/* ================ Local functions ================================================================================= */
/**
 * writes digit to file
 * @param n [in] digit
 * @param nDigits [in] length of digit
 * @param f [in] file to write to
 * @return size of bytes written
 */
static size_t fwriteDigit(DIGIT_T n[], size_t const nDigits, FILE* f)
{
    int i;
    size_t wb = 0;
    for (i = (int) nDigits - 1; i >= 0; i--) {
        char str[BYTES_PER_DIGIT * 2] = { 0 };
        (void) sprintf(str, "%08lX", (long unsigned int) n[i]);
        wb += fwrite(str, sizeof(str), 1, f);
    }
    return wb;
}

/**
 * writes digit to file as C-array
 * @param n [in] digit
 * @param nDigits [in] length of digit
 * @param f [in] file to write to
 * @return size of bytes written
 */
static size_t fwriteCarray(DIGIT_T n[], size_t const nDigits, FILE* f)
{
    unsigned char const * const byteArray = (unsigned char const * const ) n;
    int i;
    size_t wb = 0;
    for (i = (int) (nDigits * BYTES_PER_DIGIT) - 1; i >= 0; i--) {
        char str[BYTES_PER_DIGIT * 4] = { 0 };
        (void) sprintf(str, "0x%02x, ", (unsigned char) byteArray[i]);
        wb += fwrite(str, sizeof(str), 1, f);
    }
    return wb;
}

/**
 * Generates random digit between range lower and upper
 * @param lower [in] lower bound of range
 * @param upper [in] upper bound of range
 * @return random digit between range
 */
static DIGIT_T spPseudoRand(DIGIT_T lower, DIGIT_T upper)
{ /*  Returns a pseudo-random digit.
 Handles own seeding using time
 NOT for cryptographically-secure random numbers.
 */
    static unsigned seeded = 0;
    unsigned int i;
    double f;

    if (!seeded) {
        /* seed with system time */
        //srand(0);
        /* Throw away a random few to avoid similar starts */
        i = (unsigned int) rand() & 0xFFU;
        while (i--) {
            rand();
        }
        seeded = 1;
    }
    f = ((double) rand() / (double) RAND_MAX) * (double) upper;
    return (lower + (DIGIT_T) f);
}

/**
 * Generates prime number
 * @param p [out] prime number to generate
 * @param ndigits [in] length of prime number
 * @return success condition
 */
static int GenRsaPrime(DIGIT_T p[], unsigned int const ndigits)
{

    unsigned int i;

    for (i = 0; i < ndigits; i++) {
        p[i] = spPseudoRand(0, MAX_DIGIT);
    }

    // make sure two highest and the low bits are set. Having the two highest bits set means the product (pq) will always have its highet bit set
    p[ndigits - 1] |= HIBITMASK | (HIBITMASK >> 1);
    p[0] |= 0x1;

    // check if prime
    int isPrime;
    do {
        isPrime = mpIsPrime(p, ndigits, 10);
        if (isPrime == 0) {
            mpShortAdd(p, p, 2, ndigits);

            // check for overflow
            if (!(p[ndigits - 1] & HIBITMASK)) {
                // failed to find a prime
                return -1;
            }
        }
    } while (isPrime == 0);

    return 0;
}

/**
 * generate RSA keys
 * @param n [out] RSA private key
 * @param e [out] RSA public key
 * @param d [out] RSA modulus digit
 * @param modSize length if digits
 * @return succesful completion of operation
 */
static int mpGenerateRSAKeys(DIGIT_T n[], DIGIT_T e[], DIGIT_T d[], unsigned int const modSize)
{
    // generate an RSA key pair (n, e), (d) with key size MOD_SIZE digits
    DIGIT_T p[modSize], q[modSize];
    DIGIT_T L[2 * modSize];
    DIGIT_T g[modSize], t[modSize];
    unsigned int pndigits;
    unsigned int qndigits;
    int status;
    int retCode = 0;
    int cmp;

    // set e to know public key value
    mpSetZero(e, modSize);
    e[0] = 0x10001;

    // generate random prime p with half the length of the modulus
    mpSetZero(p, modSize);
    pndigits = modSize / 2;
    do {
        status = GenRsaPrime(p, pndigits);
        if (status != 0) {
            retCode = -1;
        } else {
            // make sure gcd(p-1, e) = 1
            mpShortSub(L, p, 1, pndigits);
            mpGcd(g, L, e, pndigits);
            cmp = mpShortCmp(g, 1, pndigits);
        }
    } while ((cmp != 0) && (retCode == 0));

    if (retCode == 0) {
        // generate random prime q for balance of modulus length
        mpSetZero(q, modSize);
        qndigits = modSize - pndigits;
        do {
            status = GenRsaPrime(q, qndigits);
            if (status != 0) {
                retCode = -1;
            } else {
              // make sure gcd(q-1, e) = 1
              mpShortSub(L, q, 1, qndigits);
              mpGcd(g, L, e, qndigits);
              cmp = mpShortCmp(g, 1, qndigits);
            }
        } while ((cmp != 0) && (retCode == 0));

        if (retCode == 0) {
            // calculate modulus, n = pq using L as a double-length temp
            mpMultiply(L, p, q, modSize); /* L is double-length */
            mpSetEqual(n, L, modSize); /* but n is only single */

            // set L equal to (p-1)*(q-1)
            mpShortSub(g, p, 1, modSize);
            mpShortSub(t, q, 1, modSize);
            mpMultiply(L, g, t, modSize);

            // calculate private key using the public key d = e^-1 Mod L
            mpModInv(d, e, L, modSize);

            // cleanup data
            mpSetZero(p, modSize);
            mpSetZero(q, modSize);
            mpSetZero(g, modSize);
            mpSetZero(t, modSize);
            mpSetZero(L, 2 * modSize);
        }
    }
    return retCode;
}

/**
 * Prints a help message to the standard console
 * @param doStandardHelp
 */
static void help(int const doStandardHelp)
{
    if (doStandardHelp)
        fprintf(stderr, "Try ``generate_keys --help'' for more information.\n");
    else
        printf("generate_keys (OSB " VERSION ")\n"
                        "Usage: generate_keys [RSA-OUTPUT-FILE] [RSA-LENGTH] [N_RSA_KEYS] [AES-OUTPUT-FILE] [AES-LENGTH] [N_AES_KEYS]\n"
                        "Generate RSA and AES keys with RSA/AES key length in bits.\n"
                        "\n"
                        "-q, --quiet                suppress all normal output\n"
                        "-h, --help                 display this help and exit\n"
                        "-v, --version              output version information and exit\n"
                        "\n"
                        "Report bugs to <info@omron.com>.\n");

    exit(doStandardHelp);
}

/**
 *
 * @param rsaFile
 * @param rsaLengthBits
 * @param nRsaKeys
 * @param aesFile
 * @param aesLengthBits
 * @param nAesKeys
 * @return
 */
static int generateKeys(char const * const  rsaFile,
                        int const           rsaLengthBits,
                        int const           nRsaKeys,
                        char const * const  aesFile,
                        int const           aesLengthBits,
                        int const           nAesKeys)
{
    int                status          = OMN_EXIT_SUCCESS;
    unsigned int const rsaLengthDigits = rsaLengthBits / BITS_PER_DIGIT;
    unsigned int const aesLengthDigits = aesLengthBits / BITS_PER_DIGIT;

    if (((aesLengthBits % BITS_PER_DIGIT) != 0) || ((aesLengthDigits * BYTES_PER_DIGIT) != AES128_KEYLEN)) {
        status = UNSUPPORTED_AES_KEYLENGTH;
    }
    if ((rsaLengthBits % BITS_PER_DIGIT) != 0) {
        status = UNSUPPORTED_RSA_KEYLENGTH;
    }

    if ((nRsaKeys < 0) || (nAesKeys < 0)) {
        status = PARAM_ERROR;
    }

    if (status == OMN_EXIT_SUCCESS) {
        DIGIT_T n[nRsaKeys][rsaLengthDigits];
        DIGIT_T e[nRsaKeys][rsaLengthDigits];
        DIGIT_T d[nRsaKeys][rsaLengthDigits];
        DIGIT_T aes[nAesKeys][aesLengthDigits];
        FILE*   f;
        unsigned int i;
        unsigned int idx;

        if (quiet != 1) {
            printf("generating keys...\n");
        }
        if (quiet != 1) {
            printf("AES-%d", aesLengthBits);
        }
        for (idx = 0; idx < nAesKeys; idx++) {
            for (i = 0; i < aesLengthDigits; i++) {
                aes[idx][i] = (DIGIT_T) rand();
            }
            if (quiet != 1) {
                printf(".");
            }
        }
        if (quiet != 1) {
            printf("\nRSA-%d", rsaLengthBits);
        }
        for (idx = 0; idx < nRsaKeys; idx++) {
            status |= mpGenerateRSAKeys(n[idx], e[idx], d[idx], rsaLengthDigits);
            if (quiet != 1) {
                printf(".");
            }
        }
        if (quiet != 1) {
            printf("\n");
        }

        if (status == OMN_EXIT_SUCCESS) {
            if (quiet != 1) {
                for (idx = 0; idx < nRsaKeys; idx++) {
                    printf("rsa-%i\n", rsaLengthDigits * BITS_PER_DIGIT);
                    printf("n (modulus)\n");
                    mpPrintTrimNL(n, rsaLengthDigits);
                    printf("e (public encryption/verifying)\n");
                    mpPrintTrimNL(e, rsaLengthDigits);
                    printf("d (private decryption/signing)\n");
                    mpPrintTrimNL(d, rsaLengthDigits);
                }
                for (idx = 0; idx < nAesKeys; idx++) {
                    printf("aes-%i\n", aesLengthDigits * BITS_PER_DIGIT);
                    mpPrintTrimNL(aes[idx], aesLengthDigits);
                }
            }
            // write RSA keys in text
            f = fopen(rsaFile, "w+");
            if (f == NULL) {
                status = RSA_OUTFILE_CREATE_ERROR;
            } else {
                size_t wb = 0;
                for (idx = 0; idx < nRsaKeys; idx++) {
                    wb += fwrite("n: ", 3, 1, f);
                    wb += fwriteDigit(n[idx], rsaLengthDigits, f);
                    wb += fwrite("\ne: ", 4, 1, f);
                    wb += fwriteDigit(e[idx], rsaLengthDigits, f);
                    wb += fwrite("\nd: ", 4, 1, f);
                    wb += fwriteDigit(d[idx], rsaLengthDigits, f);
                    wb += fwrite("\n", 1, 1, f);
                }

                if (wb != (nRsaKeys * 7)) {
                    printf("wb: %d\n", wb);
                    status = RSA_OUTFILE_WRITE_ERROR;
                }

                fclose(f);
            }
            // write AES key in binary form
            f = fopen(aesFile, "w+");
            if (f == NULL) {
                status = AES_OUTFILE_CREATE_ERROR;
            } else {
                // write least significant bit first
                for (idx = 0; idx < nAesKeys; idx++) {
                    (void)misc_fwriteHex(f, (t_U8*) aes[idx], 0);
                    (void)fwrite("\n", 1, 1, f);
                }

                fclose(f);
            }
        }
    }
    return status;
}

/* ================ Global functions (Sub-system scope) ============================================================= */

/* ================ Global functions (Global system scope) ======================================================= *//**
 @endinternal
 *//* ================================================================================================================ */

/**
 * Main entry function of the application
 * @param argc [in] number of arguments
 * @param argv [in] values of arguments
 * @return successful completion
 */
int main(int argc, char *argv[])
{
    int c;

    srand((unsigned int) time(NULL));

    do {
        c = getopt_long(argc, argv, optstring, longopts, 0);
        switch (c) {
            case EOF:
                break;

            case 'h':
                help(0);
                break;

            case 'v':
                printf("generate_keys (OSB " VERSION ")\n");
                exit(0);
                break;

            case 'q':
                quiet = 1;
                break;

            default:
                help(1);
                break;
        }
    } while (c != EOF);

    if (optind < argc) {
        if (argc == REQ_ARGS_CNT) {
            char const * const rsaFile  = argv[optind++];
            int const rsaBitLength      = (int) atoi(argv[optind++]);
            int const nRsaKeys          = (int) atoi(argv[optind++]);
            char const * const aesFile  = argv[optind++];
            int const aesBitLength      = (int) atoi(argv[optind++]);
            int const nAesKeys          = (int) atoi(argv[optind++]);
            int const status            = generateKeys(rsaFile, rsaBitLength, nRsaKeys, aesFile, aesBitLength, nAesKeys);

            if (status != OMN_EXIT_SUCCESS) {
                fprintf(stderr, "ERROR: '%s'\n", errMsg[status]);
                exit(1);
            }
        } else {
            fprintf(stderr, "Please provide the required parameters: 'generate_keys [RSAFILE] [RSALEN] [N_RSA_KEYS] [AESFILE] [AESLEN] [N_AES_KEYS]'\n");
            exit(1);
        }
    } else {
    }
    return 0;
}

/* ================ End of file ===================================================================================== */
