/* =====================================================================================================================
 (c) Copyright OMRON Europe BV 2016
 All Right Reserved
 OMRON Development & Engineering
 ================================================================================================================== *//**
 @file

  This file is part of Omron Secure Boot (OSB).

  Omron Secure Boot (OSB) is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Omron Secure Boot (OSB) is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Omron Secure Boot (OSB).  If not, see <http://www.gnu.org/licenses/>.

 @internal
 *//* ================================================================================================================ */

/* ================ Global include files ============================================================================ */

#include "misc.h"
#include "aes.h"
#include "types.h"
#include <string.h>

/* ================ Sub-system include files ======================================================================== */

/* ================ Local constant definitions ====================================================================== */

/* ================ Local macro definitions ========================================================================= */

/* ================ Local type definitions ========================================================================== */

/* ================ Local data definitions ========================================================================== */

/* ================ Local functions ================================================================================= */
/**
 *
 * @param c
 * @return
 */
static char toUpper(char const c) {
    if (('a' <= c) && (c <= 'z')) {
        return 'A' + (c - 'a');
    }
    return c;
}

/**
 *
 * @param c
 * @return
 */
static t_bool isHex(char const c) {
    char cUpper = toUpper(c);

    return (t_bool) ((cUpper >= 'A' && cUpper <= 'F') || (cUpper >= '0' && cUpper <= '9')) ? TRUE : FALSE;

}

/**
 *
 * @param buf
 * @param bufLen
 */
static void swapArray(t_U8* const buf, int const bufLen) {
    int i;

    for (i = 0; i < bufLen / 2; i++) {
        t_U8 tmp = buf[i];
        buf[i] = buf[bufLen - 1 - i];
        buf[bufLen - 1 - i] = tmp;
    }
}

/**
 *
 * @param c
 * @return
 */
static t_U8 hexDigitToByte(char const c) {
    char cUpper = toUpper(c);
    t_U8 byte = 0;

    if ((cUpper >= 'A' && cUpper <= 'Z')) {
        byte = (t_U8) ((cUpper - 'A') + 10);
    } else if ((cUpper >= '0' && cUpper <= '9')) {
        byte = (t_U8) (cUpper - '0');
    }
    return byte;
}

/**
 *
 * @param s
 * @return
 */
static t_U8 hexToByte(char const * const s) {
    return (hexDigitToByte(s[0]) << 4) + hexDigitToByte(s[1]);
}

/**
 * Retrieves binary key from the provided file (represented as hex string) and stores it in an array
 * @param keyfile   [in]    The filename (path+file) to the file containing the encryption key
 * @param buf       [in]    Buffer to fill with hex-values
 * @param bufLen    [inout]  In: Maximum length of buffer; Out: size of buffer
 * @return
 */
static void freadHex(FILE* const keyfile, t_U8* const buf, int * const bufLen) {
    t_bool finished = FALSE;
    int index = 0;
    do {
        char str[2];
        size_t rb = fread(str, 1, 2, keyfile);
        if ((index >= *bufLen) || (rb != 2) || ISFALSE(isHex(str[0])) || ISFALSE(isHex(str[1]))) {
            finished = TRUE;
        }
        if (ISFALSE(finished)) {
            buf[index] = hexToByte(str);
            index++;
        }
    } while (ISFALSE(finished));
    swapArray(buf, index);
    *bufLen = index;
}

/* ================ Global functions (Sub-system scope) ============================================================= */

/* ================ Global functions (Global system scope) ======================================================= */

/**
 * Prints key in LSB
 * @param key    [in]   Key to print
 * @param keyLen [in]   Length of key
 * @return void
 */
void misc_printKey(const t_U8 const * const key, int const keyLen) {
    int i;

    for (i = keyLen - 1; i >= 0; i--) {
        printf("%02X ", ((unsigned char const * const ) key)[(unsigned) i]);
    }
    printf("\n");
}

/**
 * Prints hash in MSB
 * @param hash    [in]   HASH to print
 * @param hashLen [in]   Length of hash
 * @return void
 */
void misc_printHash(const t_U8 const * const hash, int const hashLen) {
    int i;

    for (i = 0; i < hashLen; i++) {
        printf("%02x ", ((unsigned char const * const ) hash)[(unsigned)i]);
    }
    printf("\n");
}

/**
 * Stores the binary AES key to the provided file (represented as hex string)
 * @param keyfile   [in]    The filename (path+file) to the file containing the encryption key
 * @param buf       [out]   Will receive the converted data from the file
 * @param asCarray  [in]    Output a C compatible array (set to '1')
 * @return hex-string bytes written to file (2 bytes per hex-string)
 */
int misc_fwriteHex(FILE* const keyfile, t_U8 const * const buf, t_U8 const asCarray)
{
    int i;
    int wbTotal = 0;

    if (asCarray != 0) {
        wbTotal += (int)fwrite("{ ", 2, 1, keyfile);
    }
    for (i = AES128_KEYLEN - 1; i >= 0; i--) {
        char str[4] = { 0 };

        if (asCarray != 0) {
            (void) sprintf(str, "0x%02x, ", ((unsigned char const * const ) buf)[(unsigned) i]);
        } else {
            (void) sprintf(str, "%02X", ((unsigned char const * const ) buf)[(unsigned) i]);
        }
        wbTotal += (int)fwrite(str, ((asCarray != 0) ? 4 : 2), 1, keyfile);
    }
    if (asCarray != 0) {
        wbTotal += (int)fwrite("};", 2, 1, keyfile);
    }
    return wbTotal;
}

/**
 * Retrieves the binary AES key from the provided file and stores it in an array
 * @param filename  [in]   The filename (path+file) to the file containing the encryption key
 * @param keyIndex  [in]   Key at key-file index 'keyIndex', starting from 0
 * @param key       [out]  Will receive the converted data from the file
 * @param keyLen    [inout]  In: Maximum length of buffer Out: size of buffer
 * @return 0 when no error
 */
int misc_getAesKey(char const * const filename, int const keyIndex, t_U8* const key, int * const keyLen)
{
    int status = OMN_EXIT_FAILURE;
    FILE* const keyFile = fopen(filename, "r");

    if ((keyFile != NULL) && (keyIndex >= 0)) {
        int const hasError = fseek(keyFile, keyIndex * ((AES128_KEYLEN * 2) + 1), SEEK_SET);

        if (hasError == 0) {
            freadHex(keyFile, key, keyLen);
            (void) fclose(keyFile);
            status = OMN_EXIT_SUCCESS;
        }
    }

    return status;
}

/* ================ End of file ===================================================================================== */

