/* =====================================================================================================================
 (c) Copyright OMRON Europe BV 2016
 All Right Reserved
 OMRON Development & Engineering
 ================================================================================================================== *//**
 @file

  This file is part of Omron Secure Boot (OSB).

  Omron Secure Boot (OSB) is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Omron Secure Boot (OSB) is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Omron Secure Boot (OSB).  If not, see <http://www.gnu.org/licenses/>.

 Linux command line tool to generate a SHA hash of the provided file

 @note Only SHA is supported now

 @internal
 *//* ================================================================================================================ */

/* ================ Global include files ============================================================================ */
#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <stdarg.h>
#include <time.h>
#include "sha2.h"
#include "types.h"
#include "rsa.h"
#include "rsa_cfg.h"
#include "misc.h"
#include "bigdigits.h"

/* ================ Sub-system include files ======================================================================== */

/* ================ Local constant definitions ====================================================================== */
#define SIGN_FILE_EXT           ".sig"

#define REQ_ARGS_CNT            (4)         //filename + two arguments (keyfile and file to sign)

#define SUCCESS                 (0)
#define FILE_NOT_FOUND          (1)
#define OUTFILE_CREATE_ERROR    (2)
#define OUTFILE_WRITE_ERROR     (3)
#define INTERNAL_BUF_SIZE_ERROR (4)
#define PARSE_ERROR             (5)
#define KEYFILE_NOT_FOUND       (6)
#define FILE_ERROR              (7)
#define SIGN_FAILURE            (8)

#define CMP_BUF_SIZE                (8)
#define MODULO_ENTRY_IDENTIFIER     "n: "
#define DECRYPT_ENTRY_IDENTIFIER    "d: "
#define ENCRYPT_ENTRY_IDENTIFIER    "e: "
#define CHARS_PER_BYTE              (2)

/* ================ Local macro definitions ========================================================================= */

/* ================ Local type definitions ========================================================================== */
void grub_printf(const char *format, ...);
void grub_memset(void *start, int c, int len);
int grub_strlen(const char *str);
void* grub_memmove(void *to, const void *from, int len);
int grub_isspace(int c);
int grub_memcmp(const char *s1, const char *s2, int n);
void* grub_memcpy(void *dest, const void *src, int n);

/* ================ Local data definitions ========================================================================== */
static char* errMsg[] = { [FILE_NOT_FOUND] = "Could not open the file to sign", [OUTFILE_CREATE_ERROR] = "Could not create the output file",
                          [OUTFILE_WRITE_ERROR] = "Could not write to the output file", [INTERNAL_BUF_SIZE_ERROR
                                          ] = "Internal buffers cannot hold the converted RSA keys",
                          [PARSE_ERROR] = "Something went wrong when parsing the RSA key-file", [KEYFILE_NOT_FOUND
                                          ] = "Could not find the provided key-file",
                          [FILE_ERROR] = "Error occured during file-access", [SIGN_FAILURE
                                          ] = "An error occured while signing the generated SHA hash", };

static int verbose = 0;
static char *optstring = "hv";
static struct option longopts[] = { { "help", no_argument, 0, 'h' }, { "version", no_argument, 0, 'v' }, { 0 } };

/* ================ Local functions ================================================================================= */
/**
 *
 * @param format
 */
void grub_printf(const char *format, ...)
{
    va_list args;
    va_start(args, format);

    printf(format, args);

    va_end(args);

}

/**
 *
 * @param start
 * @param c
 * @param len
 */
void grub_memset(void *start, int c, int len)
{
    (void) memset(start, c, len);
}

/**
 *
 * @param str
 * @return
 */
int grub_strlen(const char *str)
{
    return strlen(str);
}

/**
 *
 * @param s1
 * @param s2
 * @param n
 * @return
 */
int grub_memcmp(const char *s1, const char *s2, int n)
{
    return memcmp(s1, s2, n);
}

/**
 *
 * @param to
 * @param from
 * @param len
 */
void* grub_memmove(void *to, const void *from, int len)
{
    return memcpy(to, from, len);
}

/**
 *
 * @param c
 * @return
 */
int grub_isspace(int c)
{
    switch ((char) c) {
        case ' ':
        case '\t':
        case '\r':
        case '\n':
            return 1;
        default:
            return 0;
    }
}

/**
 *
 * @param dest
 * @param src
 * @param n
 */
void* grub_memcpy(void *dest, const void *src, int n)
{
    return memcpy(dest, src, n);
}

/**
 * Prints a help message to the standard console
 * @param doStandardHelp
 */
static void help(int const doStandardHelp)
{
    if (doStandardHelp)
        fprintf(stderr, "Try ``sign_file --help'' for more information.\n");
    else
        printf("sign_file (OSB " VERSION ")\n",
                        "Usage: sign_file [KEYFILE] [KEY_IDX] [INFILE]\n"
                                        "Generates a " SIGN_FILE_EXT "-file from the input file [INFILE] and the private RSA key-file ([KEYFILE]) with key-index [KEY_IDX] and writes the result to '[INFILE]" SIGN_FILE_EXT "'.\n"
                                        "\n"
                                        "-h, --help                 display this help and exit\n"
                                        "-v, --version              output version information and exit.\n"
                                        "\n"
                                        "Report bugs to <info@omron.com>.\n");

    exit(doStandardHelp);
}

/**
 *
 * @param keyFile
 * @return
 */
static int nextLine(FILE* const keyFile)
{
    int bytesRead;
    int found = 0;

    do {
        char chr;

        bytesRead = fread(&chr, 1, 1, keyFile);
        if ((bytesRead != 0) && ((chr == '\n') || (chr == '\r'))) {
            found = 1;
        }
    } while ((bytesRead != 0) && (found == 0));

    if (found != 0) {
        char chr;

        do {
            bytesRead = fread(&chr, 1, 1, keyFile);
        } while ((bytesRead != 0) && ((chr == '\n') || (chr == '\r')));
        fseek(keyFile, -1, SEEK_CUR); //Oeps, we read the first normal byte of the file. Correct for this so file-pointer is just after \n or \r
    }

    return (int)((found != 0) ? OMN_EXIT_SUCCESS : OMN_EXIT_FAILURE);
}

/**
 *
 * @param keyFile
 * @param keyIndex
 * @param locateStr
 * @param locStrLen
 * @param output
 * @return
 */
static int locateAndConvert(FILE* const keyFile, int const keyIndex, char const * const locateStr, size_t const locStrLen, DIGIT_T* const output)
{
    int fstatus;
    int success = PARSE_ERROR;
    size_t bytesCopied = 0;
    size_t bytesRead;
    int isDifferent = 1;
    char rsaBuf[(RSA_KEYLEN * CHARS_PER_BYTE) + 1]; //+1 for NUL-term
    char* rsaBufIdx = rsaBuf;
    char buffer[locStrLen];
    int  actualKeyIndex = -1;

    fstatus = fseek(keyFile, 0UL, SEEK_SET);
    if (fstatus == 0) {
        do {
            bytesRead = fread(buffer, 1, locStrLen, keyFile);
            isDifferent = memcmp(buffer, locateStr, locStrLen);

            while ((isDifferent != 0) && (bytesRead != 0)) {
                int const status = nextLine(keyFile);
                if (status == OMN_EXIT_SUCCESS) {
                    bytesRead = fread(buffer, 1, locStrLen, keyFile);
                    isDifferent = memcmp(buffer, locateStr, locStrLen);
                } else {
                    isDifferent = 1;
                    bytesRead   = 0;
                }
            }
            actualKeyIndex++;
        } while ((actualKeyIndex < keyIndex) && (isDifferent == 0));

        if ((isDifferent == 0) && (bytesRead == locStrLen)) {
            int readError = 0;
            do {
                bytesRead = fread(rsaBufIdx, 1, 1, keyFile);
                if ((bytesRead == 1) && (*rsaBufIdx != '\n') && (*rsaBufIdx != '\r')) {
                    rsaBufIdx++;
                    bytesCopied++;
                } else {
                    readError = 1;
                }
            } while ((bytesCopied < sizeof(rsaBuf)) && (readError == 0));

            if (bytesCopied == sizeof(rsaBuf) - 1) {
                rsaBuf[bytesCopied] = 0;
                (void) mpConvFromHex(output, RSA_KEYLEN / BYTES_PER_DIGIT, rsaBuf);
                success = SUCCESS;
            } else {
                success = PARSE_ERROR;
            }
        }
    } else {
        success = FILE_ERROR;
    }
    return success;
}

/**
 * Retrieves the private RSA sign key from the provided file and stores it in an array
 * @param keyfile   [in]    The filename (path+file) to the file containing the private RSA sign key
 * @param keyIndex  [in]
 * @param signKey   [out]
 * @param signSize  [in]
 * @param modulo    [out]
 * @param moduloSize[in]
 * @return 0 when no error
 */
static int getSignKey(char const * const keyfile, int const keyIndex, DIGIT_T* const signKey, int const signSize, DIGIT_T* const modulo, int const moduloSize)
{
    int status = PARSE_ERROR;
    FILE* const keyFile = fopen(keyfile, "r");

    if ((keyFile != NULL) && (keyIndex >= 0)) {
        if (((RSA_KEYLEN / BYTES_PER_DIGIT) <= moduloSize) && ((RSA_KEYLEN / BYTES_PER_DIGIT) <= signSize)) {
            status = locateAndConvert(keyFile, keyIndex, MODULO_ENTRY_IDENTIFIER, sizeof(MODULO_ENTRY_IDENTIFIER) - 1, modulo);
            if (status == SUCCESS) {
                status = locateAndConvert(keyFile, keyIndex, DECRYPT_ENTRY_IDENTIFIER, sizeof(DECRYPT_ENTRY_IDENTIFIER) - 1, signKey);
            }
            (void) fclose(keyFile);
        } else {
            status = INTERNAL_BUF_SIZE_ERROR;
        }
    } else {
        status = KEYFILE_NOT_FOUND;
    }

    return status;
}

/**
 * Generates the SHA HASH file from the input file
 * @param inFile    [in]  File stream of the input file
 * @param modulo    [in]  Required modulo for the RSA decrypt (sign) function
 * @param signKey   [in]  RSA private key to sign the hash output
 * @param outFile   [out] File stream to which the HASH data is written to
 * @return 0 when successful
 */
static int signData(FILE* const inFile, DIGIT_T* const signKey, DIGIT_T* const modulo, FILE* const outFile)
{
    SHA512_CTX shaCtx;
    int bytesRead;
    int bytesWritten;
    int status = SUCCESS;
    t_U8 shaBuf[SHA_BUF_SIZE];
    t_U32 shaResult[SHA2_DWORD_SIZE];
    int outputLen = RSA_KEYLEN;
    t_U8 output[outputLen];

    SHA512_Init(&shaCtx);
    (void) memset(shaBuf, 0x00, sizeof(shaBuf));

    bytesRead = (int) fread(shaBuf, 1, (size_t) sizeof(shaBuf), inFile);
    while (bytesRead == (int) sizeof(shaBuf)) {
        SHA512_Update(&shaCtx, shaBuf, bytesRead);
        bytesRead = (int) fread(shaBuf, 1, (size_t) sizeof(shaBuf), inFile);
    }
    if (bytesRead > 0) {
        (void) SHA512_Update(&shaCtx, shaBuf, bytesRead);
    }
    SHA512_Final((sha2_byte*) shaResult, &shaCtx);

    status = rsa_create(modulo, signKey, (t_U8*) shaResult, sizeof(shaResult), output, &outputLen);
    if (status == OMN_EXIT_SUCCESS) {
        if (verbose == 1) {
            printf("Private key: ");
            misc_printKey((t_U8 const* const ) signKey, RSA_KEYLEN);
            printf("Modulus: ");
            misc_printKey((t_U8 const* const ) modulo, RSA_KEYLEN);
            printf("Hash: ");
            misc_printHash((t_U8 const* const ) shaResult, sizeof(shaResult));
        }

        bytesWritten = (int) fwrite(output, 1, (unsigned) outputLen, outFile);
        if (bytesWritten == outputLen) {
            status = SUCCESS;
        } else {
            status = OUTFILE_WRITE_ERROR;
        }
    } else {
        status = SIGN_FAILURE;
    }

    return status;
}

/**
 * Preparation function to hash an input file and generate the output file
 * @param keyfile   [in] Filename of the RSA key-file (including path)
 * @param keyIndex  [in] The key to use from the keyfile, starting at 0
 * @param infile    [in] Filename of the input-file (including path)
 * @return 0 when successful
 */
static int signFile(char const * const keyfile, int const keyIndex, char const * const infile)
{
    char outfile[strlen(infile) + strlen(SIGN_FILE_EXT)];
    int status;
    FILE* inFile;
    FILE* outFile;
    DIGIT_T signKey[RSA_KEYLEN / BYTES_PER_DIGIT];
    DIGIT_T modulo[RSA_KEYLEN / BYTES_PER_DIGIT];

    strcpy(outfile, infile);
    (void) strcat(outfile, SIGN_FILE_EXT);
    outFile = fopen(outfile, "w+");
    if (outFile == NULL) {
        status = OUTFILE_CREATE_ERROR;
        goto error;
    }

    status = getSignKey(keyfile, keyIndex, signKey, sizeof(signKey), modulo, sizeof(modulo));
    if (status != SUCCESS) {
        (void) fclose(outFile);
        goto error;
    }

    inFile = fopen(infile, "r");
    if (inFile == NULL) {
        (void) fclose(outFile);
        status = FILE_NOT_FOUND;
        goto error;
    }

    status = signData(inFile, signKey, modulo, outFile);
    if (status != 0) {
        (void) fclose(inFile);
        (void) fclose(outFile);
        goto error;
    } else {
        printf("Created signed hash '%s%s'\n", infile, SIGN_FILE_EXT);
    }

    (void) fclose(inFile);
    (void) fclose(outFile);

    error : return status;
}

/* ================ Global functions (Sub-system scope) ============================================================= */

/* ================ Global functions (Global system scope) ======================================================= *//**
 @endinternal
 *//* ================================================================================================================ */
/**
 * Main entry function of the application
 * @param argc
 * @param argv
 * @return
 */
int main(int argc, char *argv[])
{
    int c;

    srand((unsigned int) time(NULL));

    do {
        c = getopt_long(argc, argv, optstring, longopts, 0);
        switch ((char) c) {
            case (char) EOF:
                break;

            case 'h':
                help(0);
                break;

            case 'v':
                printf("sign_file (OSB " VERSION ")\n");
                verbose = 1;
                break;

            default:
                help(1);
                break;
        }
    } while (c != EOF);

    if (optind < argc) {
        if (((verbose == 0) && (argc == REQ_ARGS_CNT)) || ((verbose == 1) && (argc == REQ_ARGS_CNT + 1))) {
            char const * const keyfile = argv[optind++];
            int const keyIndex         = (int) atoi(argv[optind++]);
            char const * const infile  = argv[optind++];
            int const status = signFile(keyfile, keyIndex, infile);

            if (status != 0) {
                (void) fprintf(stderr, "ERROR: '%s'\n", errMsg[status]);
                exit(1);
            }
        } else {
            (void) fprintf(stderr, "Please provide the required parameters: 'sign_file [KEYFILE] [KEY_IDX] [INFILE]'\n");
            exit(1);
        }
    } else {
    }

    return 0;
}

/* ================ End of file ===================================================================================== */
